from flo_ai.arium.protocols import ExecutableNode
from typing import List, Any, Dict, Optional, TYPE_CHECKING
from flo_ai.utils.logger import logger
from flo_ai.arium.memory import MessageMemory

if TYPE_CHECKING:  # need to have an optional import else will get circular dependency error as arium also has AriumNode reference
    from flo_ai.arium.arium import Arium


class AriumNode:
    """
    Wrapper to use an Arium as a node in another Arium workflow.
    """

    def __init__(self, name: str, arium: 'Arium', inherit_variables: bool = True):
        """
        Args:
            name: Name for this node in the parent workflow
            arium: The Arium workflow to execute
            inherit_variables: Whether to pass parent variables to sub-workflow
        """
        self.name = name
        self.arium = arium
        self.inherit_variables = inherit_variables

    async def run(
        self, inputs: List[Any], variables: Optional[Dict[str, Any]] = None, **kwargs
    ) -> Any:
        """Execute the nested Arium workflow with isolated memory"""

        # Handle variable inheritance
        execution_variables = (
            variables.copy() if (self.inherit_variables and variables) else None
        )

        # Execute the nested Arium with isolated memory
        result = await self.arium.run(
            inputs=inputs,
            variables=execution_variables,
        )
        return result


class ForEachNode:
    """
    Execute a node on each item in a collection.

    Supports only sequential execution for now. (parallel execution would be supported in future)
    """

    def __init__(self, name: str, execute_node: ExecutableNode):
        """
        Args:
            name: Node name
            execute_node: Node to execute on each item
        """
        self.name = name
        self.execute_node = execute_node

    async def _execute_item(
        self,
        item: Any,
        index: int,
        variables: Optional[Dict[str, Any]] = None,
    ) -> Any:
        """Execute the node on a single item"""
        logger.info(f"ForEach '{self.name}': Processing item {index + 1}")

        # Create execution variables with item context
        item_variables = (variables or {}).copy()

        # Execute the node
        result = await self.execute_node.run(
            inputs=[item],
            variables=item_variables,
        )

        # Return last item if result is a list, otherwise return as-is
        if isinstance(result, list) and result:
            return result[-1]
        return result

    async def run(
        self, inputs: List[Any], variables: Optional[Dict[str, Any]] = None, **kwargs
    ) -> List[Any]:
        """Execute the node on all items"""

        # Sequential execution
        results = []
        for i, item in enumerate(inputs):
            result = await self._execute_item(item, i, variables)
            results.append(result)

        logger.info(f"ForEach '{self.name}': Completed processing {len(results)} items")

        return results

    async def _execute_item_with_isolated_memory(
        self,
        item: Any,
        index: int,
        variables: Optional[Dict[str, Any]] = None,
    ) -> Any:
        """
        Execute the node on a single item with isolated memory.
        This prevents memory accumulation across iterations.
        """
        logger.info(
            f"ForEach '{self.name}': Processing item {index + 1} with isolated memory"
        )

        # Create execution variables with item context
        item_variables = (variables or {}).copy()

        # If the execute_node is an AriumNode, we can create a new memory instance
        if hasattr(self.execute_node, 'arium') and hasattr(
            self.execute_node.arium, 'memory'
        ):
            # Create a new memory instance for this iteration
            original_memory = self.execute_node.arium.memory
            self.execute_node.arium.memory = MessageMemory()

            try:
                # Execute the node with isolated memory
                result = await self.execute_node.run(
                    inputs=[item],
                    variables=item_variables,
                )
            finally:
                # Restore original memory
                self.execute_node.arium.memory = original_memory
        else:
            # For non-Arium nodes, execute normally
            result = await self.execute_node.run(
                inputs=[item],
                variables=item_variables,
            )

        # Return last item if result is a list, otherwise return as-is
        if isinstance(result, list) and result:
            return result[-1]
        return result



from typing import Dict, Any, List, AsyncIterator, Optional
from openai import AsyncOpenAI
from .base_llm import BaseLLM, ImageMessage
from flo_ai.tool.base_tool import Tool
from flo_ai.telemetry.instrumentation import (
    trace_llm_call,
    trace_llm_stream,
    llm_metrics,
    add_span_attributes,
)
from flo_ai.telemetry import get_tracer
from opentelemetry import trace


class OpenAI(BaseLLM):
    def __init__(
        self,
        model='gpt-4o-mini',
        api_key: str = None,
        temperature: float = 0.7,
        base_url: str = None,
        custom_headers: Optional[Dict[str, str]] = None,
        **kwargs,
    ):
        super().__init__(
            model=model, api_key=api_key, temperature=temperature, **kwargs
        )
        # Add custom headers if base_url is provided (proxy scenario)
        client_kwargs = {'api_key': api_key, 'base_url': base_url}
        if base_url and custom_headers:
            client_kwargs['default_headers'] = custom_headers

        self.client = AsyncOpenAI(**client_kwargs)
        self.model = model
        self.kwargs = kwargs

    @trace_llm_call(provider='openai')
    async def generate(
        self,
        messages: list[dict],
        functions: Optional[List[Dict[str, Any]]] = None,
        output_schema: dict = None,
        **kwargs,
    ) -> Any:
        # Handle structured output vs tool calling
        # Priority: output_schema takes precedence over functions for structured output
        if output_schema:
            # Convert output_schema to OpenAI format for structured output
            kwargs['response_format'] = {'type': 'json_object'}
            kwargs['functions'] = [
                {
                    'name': output_schema.get('title', 'default'),
                    'parameters': output_schema.get('schema', output_schema),
                }
            ]
            kwargs['function_call'] = {'name': output_schema.get('title', 'default')}

            # Add JSON format instruction to the system prompt
            if messages and messages[0]['role'] == 'system':
                messages[0]['content'] = (
                    messages[0]['content']
                    + '\n\nPlease provide your response in JSON format according to the specified schema.'
                )
            else:
                messages.insert(
                    0,
                    {
                        'role': 'system',
                        'content': 'Please provide your response in JSON format according to the specified schema.',
                    },
                )
        elif functions:
            # Use functions for tool calling when output_schema is not provided
            kwargs['functions'] = functions

        # Prepare OpenAI API parameters
        openai_kwargs = {
            'model': self.model,
            'messages': messages,
            'temperature': self.temperature,
            **self.kwargs,
            **kwargs,
        }

        # Make the API call
        response = await self.client.chat.completions.create(**openai_kwargs)
        message = response.choices[0].message

        # Record token usage if available
        if hasattr(response, 'usage') and response.usage:
            usage = response.usage
            llm_metrics.record_tokens(
                total_tokens=usage.total_tokens,
                prompt_tokens=usage.prompt_tokens,
                completion_tokens=usage.completion_tokens,
                model=self.model,
                provider='openai',
            )

            # Add token info to current span
            tracer = get_tracer()
            if tracer:
                current_span = trace.get_current_span()
                add_span_attributes(
                    current_span,
                    {
                        'llm.tokens.prompt': usage.prompt_tokens,
                        'llm.tokens.completion': usage.completion_tokens,
                        'llm.tokens.total': usage.total_tokens,
                    },
                )

        # Return the full message object instead of just the content
        return message

    @trace_llm_stream(provider='openai')
    async def stream(
        self,
        messages: List[Dict[str, Any]],
        functions: Optional[List[Dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> AsyncIterator[Dict[str, Any]]:
        """Stream partial responses from OpenAI Chat Completions API."""
        # Prepare OpenAI API parameters
        openai_kwargs = {
            'model': self.model,
            'messages': messages,
            'temperature': self.temperature,
            'stream': True,
            **self.kwargs,
            **kwargs,
        }

        if functions:
            openai_kwargs['functions'] = functions

        # Stream the API call and yield content deltas
        response = await self.client.chat.completions.create(**openai_kwargs)
        async for chunk in response:
            choices = getattr(chunk, 'choices', []) or []
            for choice in choices:
                delta = getattr(choice, 'delta', None)
                if delta is None:
                    continue
                content = getattr(delta, 'content', None)
                if content:
                    yield {'content': content}

    def get_message_content(self, response: Dict[str, Any]) -> str:
        # Handle both string responses and message objects
        if isinstance(response, str):
            return response
        # Otherwise return content if available
        return response.content if hasattr(response, 'content') else str(response)

    def format_tool_for_llm(self, tool: 'Tool') -> Dict[str, Any]:
        """Format a single tool for OpenAI's API"""
        return {
            'name': tool.name,
            'description': tool.description,
            'parameters': {
                'type': 'object',
                'properties': {
                    name: {
                        'type': info['type'],
                        'description': info['description'],
                        **(
                            {'items': info['items']}
                            if info.get('type') == 'array' and 'items' in info
                            else {}
                        ),
                    }
                    for name, info in tool.parameters.items()
                },
                'required': list(tool.parameters.keys()),
            },
        }

    def format_tools_for_llm(self, tools: List['Tool']) -> List[Dict[str, Any]]:
        """Format tools for OpenAI's API"""
        return [self.format_tool_for_llm(tool) for tool in tools]

    def format_image_in_message(self, image: ImageMessage) -> str:
        """Format a image in the message"""
        raise NotImplementedError('Not implemented image for LLM OpenAI')



from typing import Dict, Any, Optional
from .base_tool import Tool, ToolExecutionError
from flo_ai.utils.logger import logger


class PartialTool(Tool):
    """
    A tool that has some parameters pre-filled during agent building.
    The AI can still provide additional parameters during execution.
    """

    def __init__(
        self,
        base_tool: Tool,
        pre_filled_params: Dict[str, Any],
        name_override: Optional[str] = None,
        description_override: Optional[str] = None,
    ):
        """
        Create a partial tool with pre-filled parameters.

        Args:
            base_tool: The original tool to wrap
            pre_filled_params: Parameters to pre-fill (datasource_id, etc.)
            name_override: Optional custom name for the partial tool
            description_override: Optional custom description
        """
        self.base_tool = base_tool
        self.pre_filled_params = pre_filled_params.copy()

        # Create filtered parameters (remove pre-filled ones from AI's view)
        filtered_parameters = {}
        for param_name, param_info in base_tool.parameters.items():
            if param_name not in pre_filled_params:
                filtered_parameters[param_name] = param_info.copy()

        super().__init__(
            name=name_override or f'{base_tool.name}_partial',
            description=description_override
            or f'{base_tool.description} (with pre-configured parameters)',
            function=base_tool.function,
            parameters=filtered_parameters,
        )

    async def execute(self, **kwargs) -> Any:
        """Execute the tool with pre-filled parameters merged with AI-provided ones."""
        try:
            # Merge pre-filled params with AI-provided params
            # AI params take precedence over pre-filled ones
            merged_params = {**self.pre_filled_params, **kwargs}

            logger.info(
                f'Executing partial tool {self.name} with merged params: {merged_params}'
            )
            tool_result = await self.base_tool.function(**merged_params)
            logger.info(f'Partial tool {self.name} returned: {tool_result}')
            return tool_result
        except Exception as e:
            raise ToolExecutionError(
                f'Error executing partial tool {self.name}: {str(e)}', original_error=e
            )

    def get_original_tool(self) -> Tool:
        """Get the original tool without pre-filled parameters."""
        return self.base_tool

    def get_pre_filled_params(self) -> Dict[str, Any]:
        """Get the pre-filled parameters."""
        return self.pre_filled_params.copy()

    def add_pre_filled_param(self, key: str, value: Any) -> 'PartialTool':
        """Add or update a pre-filled parameter."""
        self.pre_filled_params[key] = value
        return self

    def remove_pre_filled_param(self, key: str) -> 'PartialTool':
        """Remove a pre-filled parameter."""
        if key in self.pre_filled_params:
            del self.pre_filled_params[key]
        return self


def create_partial_tool(tool: Tool, **pre_filled_params) -> PartialTool:
    """
    Create a partial tool with pre-filled parameters.

    Args:
        tool: The original tool
        **pre_filled_params: Parameters to pre-fill

    Returns:
        PartialTool: A tool with pre-filled parameters

    Example:
        # Original tool
        @flo_tool(description="Query BigQuery")
        async def bigquery_query(query: str, datasource_id: str, project_id: str):
            # implementation
            pass

        # Create partial tool with pre-filled datasource_id
        partial_tool = create_partial_tool(
            bigquery_query.tool,
            datasource_id="ds_123",
            project_id="my-project"
        )

        # AI only needs to provide the query
        result = await partial_tool.execute(query="SELECT * FROM users")
    """
    return PartialTool(tool, pre_filled_params)



from flo_ai.arium.base import BaseArium
from flo_ai.arium.memory import MessageMemory, BaseMemory
from flo_ai.llm.base_llm import ImageMessage
from flo_ai.models.document import DocumentMessage
from typing import List, Dict, Any, Optional, Callable
from flo_ai.models.agent import Agent
from flo_ai.tool.base_tool import Tool
from flo_ai.arium.models import StartNode, EndNode
from flo_ai.arium.events import AriumEventType, AriumEvent
from flo_ai.arium.nodes import AriumNode, ForEachNode
from flo_ai.utils.logger import logger
from flo_ai.utils.variable_extractor import (
    extract_variables_from_inputs,
    extract_agent_variables,
    validate_multi_agent_variables,
    resolve_variables,
)
from flo_ai.telemetry.instrumentation import workflow_metrics
from flo_ai.telemetry import get_tracer
from opentelemetry.trace import Status, StatusCode
import asyncio
import time


class Arium(BaseArium):
    def __init__(self, memory: BaseMemory):
        super().__init__()
        self.is_compiled = False
        self.memory = memory if memory else MessageMemory()

    def compile(self):
        self.validate_graph()
        self.is_compiled = True

    async def run(
        self,
        inputs: List[str | ImageMessage | DocumentMessage],
        variables: Optional[Dict[str, Any]] = None,
        event_callback: Optional[Callable[[AriumEvent], None]] = None,
        events_filter: Optional[List[AriumEventType]] = None,
    ):
        """
        Execute the Arium workflow with optional event monitoring.

        Args:
            inputs: Input messages for the workflow
            variables: Variable substitutions for templated prompts
            event_callback: Function to call for each event (if None, no events are emitted)
            events_filter: List of event types to listen for (defaults to all)

        Returns:
            List of workflow execution results
        """
        if not self.is_compiled:
            raise ValueError('Arium is not compiled')

        if not self.memory:
            raise ValueError('Arium has no memory')

        if not self.nodes:
            raise ValueError('Arium has no nodes')

        # Set default event filters to all event types if not specified
        if events_filter is None:
            events_filter = list(AriumEventType)

        # Emit workflow started event
        self._emit_event(AriumEventType.WORKFLOW_STARTED, event_callback, events_filter)

        # Get workflow name for telemetry
        workflow_name = getattr(self, 'name', 'unnamed_workflow')

        # Start telemetry tracing
        tracer = get_tracer()
        workflow_start_time = time.time()

        if tracer:
            with tracer.start_as_current_span(
                f'workflow.{workflow_name}',
                attributes={
                    'workflow.name': workflow_name,
                    'workflow.node_count': len(self.nodes),
                },
            ) as workflow_span:
                try:
                    # Extract and validate variables from inputs and all agents
                    self._extract_and_validate_variables(inputs, variables)

                    # Resolve variables in inputs and agent prompts
                    resolved_inputs = self._resolve_inputs(inputs, variables)
                    self._resolve_agent_prompts(variables)

                    # Execute the workflow with event support
                    result = await self._execute_graph(
                        resolved_inputs, event_callback, events_filter, variables
                    )

                    # Record successful workflow execution
                    workflow_duration_ms = (time.time() - workflow_start_time) * 1000
                    workflow_metrics.record_workflow(workflow_name, 'success')
                    workflow_metrics.record_workflow_latency(
                        workflow_duration_ms, workflow_name
                    )

                    workflow_span.set_status(Status(StatusCode.OK))
                    workflow_span.set_attribute(
                        'workflow.result.length', len(str(result))
                    )

                    # Emit workflow completed event
                    self._emit_event(
                        AriumEventType.WORKFLOW_COMPLETED, event_callback, events_filter
                    )

                    self.memory = MessageMemory()  # cleanup the graph

                    return result

                except Exception as e:
                    # Record failed workflow execution
                    workflow_duration_ms = (time.time() - workflow_start_time) * 1000
                    error_type = type(e).__name__

                    workflow_metrics.record_workflow(workflow_name, 'error')
                    workflow_metrics.record_error(workflow_name, error_type)
                    workflow_metrics.record_workflow_latency(
                        workflow_duration_ms, workflow_name
                    )

                    workflow_span.set_status(Status(StatusCode.ERROR, str(e)))
                    workflow_span.set_attribute('error.type', error_type)

                    # Emit workflow failed event
                    self._emit_event(
                        AriumEventType.WORKFLOW_FAILED,
                        event_callback,
                        events_filter,
                        error=str(e),
                    )
                    raise
        else:
            # No telemetry, execute without tracing
            try:
                # Extract and validate variables from inputs and all agents
                self._extract_and_validate_variables(inputs, variables)

                # Resolve variables in inputs and agent prompts
                resolved_inputs = self._resolve_inputs(inputs, variables)
                self._resolve_agent_prompts(variables)

                # Execute the workflow with event support
                result = await self._execute_graph(
                    resolved_inputs, event_callback, events_filter, variables
                )

                # Emit workflow completed event
                self._emit_event(
                    AriumEventType.WORKFLOW_COMPLETED, event_callback, events_filter
                )

                self.memory = MessageMemory()  # cleanup the graph

                return result

            except Exception as e:
                # Emit workflow failed event
                self._emit_event(
                    AriumEventType.WORKFLOW_FAILED,
                    event_callback,
                    events_filter,
                    error=str(e),
                )
                raise

    def _emit_event(
        self,
        event_type: AriumEventType,
        callback: Optional[Callable[[AriumEvent], None]],
        events_filter: Optional[List[AriumEventType]],
        **kwargs,
    ) -> None:
        """
        Emit an event if callback is provided and event type is in filtered list.

        Args:
            event_type: The type of event to emit
            callback: Function to call with the event (if None, no event is emitted)
            events_filter: List of event types to listen for
            **kwargs: Additional event data (node_name, error, etc.)
        """
        if callback and event_type in events_filter:
            event = AriumEvent(event_type=event_type, timestamp=time.time(), **kwargs)
            callback(event)

    async def _execute_graph(
        self,
        inputs: List[str | ImageMessage | DocumentMessage],
        event_callback: Optional[Callable[[AriumEvent], None]] = None,
        events_filter: Optional[List[AriumEventType]] = None,
        variables: Optional[Dict[str, Any]] = None,
    ):
        [self.memory.add(msg) for msg in inputs]

        current_node = self.nodes[self.start_node_name]
        current_edge = self.edges[self.start_node_name]

        # Loop prevention: track execution steps and node visits
        max_iterations = 20  # Reasonable limit to prevent infinite loops
        iteration_count = 0
        node_visit_count = {}  # Track how many times each node is visited
        execution_path = []  # Track the path for debugging

        logger.info(f'Executing graph from {current_node.name}')
        while current_node.name not in self.end_node_names:
            # Check for iteration limit
            iteration_count += 1
            if iteration_count > max_iterations:
                logger.error(
                    f"Maximum iterations ({max_iterations}) exceeded. Execution path: {' -> '.join(execution_path)}"
                )
                raise RuntimeError(
                    f'Workflow exceeded maximum iterations ({max_iterations}). Possible infinite loop detected.'
                )

            # Track node visits
            node_visit_count[current_node.name] = (
                node_visit_count.get(current_node.name, 0) + 1
            )
            execution_path.append(current_node.name)

            # Check for excessive node visits (same node visited too many times)
            if node_visit_count[current_node.name] > 3:
                logger.error(
                    f"Node '{current_node.name}' visited {node_visit_count[current_node.name]} times. Execution path: {' -> '.join(execution_path)}"
                )
                raise RuntimeError(
                    f"Node '{current_node.name}' visited too many times ({node_visit_count[current_node.name]}). Possible infinite loop detected."
                )

            logger.info(
                f'Executing node: {current_node.name} (iteration {iteration_count})'
            )
            # execute current node
            result = await self._execute_node(
                current_node, event_callback, events_filter, variables
            )

            if isinstance(result, List):  # for each node will give results array
                for item in result:
                    # update each item in result to memory
                    self._add_to_memory(item)
            else:
                # update results to memory
                self._add_to_memory(result)

            # find next node post current node
            # Prepare execution context for router functions
            execution_context = {
                'node_visit_count': node_visit_count,
                'execution_path': execution_path,
                'iteration_count': iteration_count,
                'current_node': current_node.name,
            }

            # Handle both sync and async router functions
            # Try to call with execution context, fallback to memory only
            try:
                router_result = current_edge.router_fn(
                    memory=self.memory, execution_context=execution_context
                )
            except TypeError:
                # Router function doesn't accept execution_context parameter
                router_result = current_edge.router_fn(memory=self.memory)

            if asyncio.iscoroutine(router_result):
                next_node_name = await router_result
            else:
                next_node_name = router_result

            # Emit router decision event
            self._emit_event(
                AriumEventType.ROUTER_DECISION,
                event_callback,
                events_filter,
                node_name=current_node.name,
                router_choice=next_node_name,
            )

            # Emit edge traversed event
            self._emit_event(
                AriumEventType.EDGE_TRAVERSED,
                event_callback,
                events_filter,
                node_name=current_node.name,
            )

            # find next edge
            # TODO: next_node_name might not be in self.edges if it's the end node. Handle this case
            next_edge = (
                self.edges[next_node_name] if next_node_name in self.edges else None
            )

            # update current node
            current_node = self.nodes[next_node_name]
            current_edge = next_edge

        return self.memory.get()

    def _extract_and_validate_variables(
        self,
        inputs: List[str | ImageMessage | DocumentMessage],
        variables: Dict[str, Any],
    ) -> None:
        """Extract variables from inputs and agents, then validate them.

        Args:
            inputs: List of input messages
            variables: Dictionary of variable name to value mappings

        Raises:
            ValueError: If any required variables are missing
        """
        # Extract variables from inputs
        input_variables = extract_variables_from_inputs(inputs)

        # Extract variables from all agents in the workflow
        agents_variables = {}
        for node in self.nodes.values():
            if isinstance(node, Agent):
                agent_vars = extract_agent_variables(node)
                if agent_vars:
                    agents_variables[node.name] = agent_vars

        # Validate input variables separately with cleaner error message
        if input_variables:
            missing_input_vars = input_variables - set(variables.keys())
            if missing_input_vars:
                provided_keys = sorted(variables.keys())
                raise ValueError(
                    f'Input contains missing variables: {sorted(missing_input_vars)}. '
                    f'Provided variables: {provided_keys}'
                )

        # Validate agent variables with detailed agent breakdown
        if agents_variables:
            validate_multi_agent_variables(agents_variables, variables)

    def _resolve_inputs(
        self,
        inputs: List[str | ImageMessage | DocumentMessage],
        variables: Dict[str, Any],
    ) -> List[str | ImageMessage | DocumentMessage]:
        """Resolve variables in input messages.

        Args:
            inputs: List of input messages
            variables: Dictionary of variable name to value mappings

        Returns:
            List of inputs with variables resolved
        """
        resolved_inputs = []
        for input_item in inputs:
            if isinstance(input_item, str):
                # Resolve variables in text input
                resolved_input = resolve_variables(input_item, variables)
                resolved_inputs.append(resolved_input)
            else:
                # ImageMessage and DocumentMessage objects don't need variable resolution
                resolved_inputs.append(input_item)
        return resolved_inputs

    def _resolve_agent_prompts(self, variables: Dict[str, Any]) -> None:
        """Resolve variables in all agent system prompts and mark them as resolved.

        Args:
            variables: Dictionary of variable name to value mappings
        """
        for node in self.nodes.values():
            if isinstance(node, Agent):
                node.system_prompt = resolve_variables(node.system_prompt, variables)
                node.resolved_variables = True

    async def _execute_node(
        self,
        node: Agent | Tool | StartNode | EndNode,
        event_callback: Optional[Callable[[AriumEvent], None]] = None,
        events_filter: Optional[List[AriumEventType]] = None,
        variables: Optional[Dict[str, Any]] = None,
    ):
        """
        Execute a single node with optional event emission.

        Args:
            node: The node to execute
            event_callback: Function to call for events (if None, no events are emitted)
            events_filter: List of event types to listen for

        Returns:
            The result of node execution
        """
        # Determine node type for events
        if isinstance(node, Agent):
            node_type = 'agent'
        elif isinstance(node, Tool):
            node_type = 'tool'
        elif isinstance(node, ForEachNode):
            node_type = 'foreach'
        elif isinstance(node, AriumNode):
            node_type = 'arium'
        elif isinstance(node, StartNode):
            node_type = 'start'
        elif isinstance(node, EndNode):
            node_type = 'end'
        else:
            node_type = 'unknown'

        # Emit node started event
        self._emit_event(
            AriumEventType.NODE_STARTED,
            event_callback,
            events_filter,
            node_name=node.name,
            node_type=node_type,
        )

        start_time = time.time()
        workflow_name = getattr(self, 'name', 'unnamed_workflow')

        # Start node telemetry tracing
        tracer = get_tracer()

        if tracer and node_type not in ['start', 'end']:
            with tracer.start_as_current_span(
                f'workflow.node.{node.name}',
                attributes={
                    'workflow.name': workflow_name,
                    'node.name': node.name,
                    'node.type': node_type,
                },
            ) as node_span:
                try:
                    # Execute the node based on its type
                    if isinstance(node, Agent):
                        # Variables are already resolved, pass empty dict to avoid re-processing
                        result = await node.run(self.memory.get(), variables={})
                    elif isinstance(node, Tool):
                        result = await node.run(inputs=[], variables=None)
                    elif isinstance(node, ForEachNode):
                        result = await node.run(
                            inputs=self.memory.get(),
                            variables=variables,
                        )
                    elif isinstance(node, AriumNode):
                        # AriumNode execution
                        result = await node.run(
                            inputs=self.memory.get(), variables=variables
                        )
                    elif isinstance(node, StartNode):
                        result = None
                    elif isinstance(node, EndNode):
                        result = None
                    else:
                        result = None

                    # Calculate execution time
                    execution_time = time.time() - start_time
                    execution_time_ms = execution_time * 1000

                    # Record node metrics
                    workflow_metrics.record_node(
                        workflow_name, node.name, node_type, 'success'
                    )
                    workflow_metrics.record_node_latency(
                        execution_time_ms, workflow_name, node.name, node_type
                    )

                    node_span.set_status(Status(StatusCode.OK))
                    node_span.set_attribute('node.execution_time_ms', execution_time_ms)

                    # Emit node completed event
                    self._emit_event(
                        AriumEventType.NODE_COMPLETED,
                        event_callback,
                        events_filter,
                        node_name=node.name,
                        node_type=node_type,
                        execution_time=execution_time,
                    )

                    return result

                except Exception as e:
                    # Calculate execution time even on failure
                    execution_time = time.time() - start_time
                    execution_time_ms = execution_time * 1000
                    error_type = type(e).__name__

                    # Record node failure
                    workflow_metrics.record_node(
                        workflow_name, node.name, node_type, 'error'
                    )
                    workflow_metrics.record_node_latency(
                        execution_time_ms, workflow_name, node.name, node_type
                    )

                    node_span.set_status(Status(StatusCode.ERROR, str(e)))
                    node_span.set_attribute('error.type', error_type)
                    node_span.set_attribute('node.execution_time_ms', execution_time_ms)

                    # Emit node failed event
                    self._emit_event(
                        AriumEventType.NODE_FAILED,
                        event_callback,
                        events_filter,
                        node_name=node.name,
                        node_type=node_type,
                        execution_time=execution_time,
                        error=str(e),
                    )

                    # Re-raise the exception
                    raise e
        else:
            # No telemetry or start/end node, execute without tracing
            try:
                # Execute the node based on its type
                if isinstance(node, Agent):
                    result = await node.run(self.memory.get(), variables={})
                elif isinstance(node, Tool):
                    result = await node.run(inputs=[], variables=None)
                elif isinstance(node, ForEachNode):
                    result = await node.run(
                        inputs=self.memory.get(),
                        variables=variables,
                    )
                elif isinstance(node, AriumNode):
                    result = await node.run(
                        inputs=self.memory.get(), variables=variables
                    )
                elif isinstance(node, StartNode):
                    result = None
                elif isinstance(node, EndNode):
                    result = None
                else:
                    result = None

                # Calculate execution time
                execution_time = time.time() - start_time

                # Emit node completed event
                self._emit_event(
                    AriumEventType.NODE_COMPLETED,
                    event_callback,
                    events_filter,
                    node_name=node.name,
                    node_type=node_type,
                    execution_time=execution_time,
                )

                return result

            except Exception as e:
                # Calculate execution time even on failure
                execution_time = time.time() - start_time

                # Emit node failed event
                self._emit_event(
                    AriumEventType.NODE_FAILED,
                    event_callback,
                    events_filter,
                    node_name=node.name,
                    node_type=node_type,
                    execution_time=execution_time,
                    error=str(e),
                )

                # Re-raise the exception
                raise e

    def _add_to_memory(self, result: str):
        # TODO result will be None for start and end nodes
        if result:
            self.memory.add(result)



from typing import Dict, Any, List, Optional, AsyncIterator
from anthropic import AsyncAnthropic
import json
from .base_llm import BaseLLM, ImageMessage
from flo_ai.tool.base_tool import Tool
from flo_ai.telemetry.instrumentation import (
    trace_llm_call,
    trace_llm_stream,
    llm_metrics,
    add_span_attributes,
)
from flo_ai.telemetry import get_tracer
from opentelemetry import trace


class Anthropic(BaseLLM):
    def __init__(
        self,
        model: str = 'claude-3-5-sonnet-20240620',
        temperature: float = 0.7,
        api_key: Optional[str] = None,
        base_url: str = None,
        custom_headers: Optional[Dict[str, str]] = None,
        **kwargs,
    ):
        super().__init__(model, api_key, temperature, **kwargs)
        # Add custom headers if base_url is provided (proxy scenario)
        client_kwargs = {'api_key': self.api_key, 'base_url': base_url}
        if base_url and custom_headers:
            client_kwargs['default_headers'] = custom_headers

        self.client = AsyncAnthropic(**client_kwargs)

    @trace_llm_call(provider='anthropic')
    async def generate(
        self,
        messages: List[Dict[str, str]],
        functions: Optional[List[Dict[str, Any]]] = None,
        output_schema: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        # Convert messages to Claude format
        system_message = next(
            (msg['content'] for msg in messages if msg['role'] == 'system'), None
        )

        # If output schema is provided, append it to system message
        if output_schema and system_message:
            system_message = f'{system_message}\n\nProvide output in the following JSON schema:\n{json.dumps(output_schema, indent=2)}\n\nResponse:'
        elif output_schema:
            system_message = f'Provide output in the following JSON schema:\n{json.dumps(output_schema, indent=2)}\n\nResponse:'

        conversation = []
        for msg in messages:
            if msg['role'] != 'system':
                conversation.append(
                    {
                        'role': 'assistant' if msg['role'] == 'assistant' else 'user',
                        'content': msg['content'],
                    }
                )

        try:
            anthropic_kwargs = {
                'model': self.model,
                'messages': conversation,
                'temperature': self.temperature,
                'max_tokens': self.kwargs.get('max_tokens', 1024),
                **self.kwargs,
                **kwargs,
            }

            if system_message:
                anthropic_kwargs['system'] = system_message

            if functions:
                anthropic_kwargs['tools'] = functions

            response = await self.client.messages.create(**anthropic_kwargs)

            # Record token usage if available
            if hasattr(response, 'usage') and response.usage:
                usage = response.usage
                llm_metrics.record_tokens(
                    total_tokens=usage.input_tokens + usage.output_tokens,
                    prompt_tokens=usage.input_tokens,
                    completion_tokens=usage.output_tokens,
                    model=self.model,
                    provider='anthropic',
                )

                # Add token info to current span
                tracer = get_tracer()
                if tracer:
                    current_span = trace.get_current_span()
                    add_span_attributes(
                        current_span,
                        {
                            'llm.tokens.prompt': usage.input_tokens,
                            'llm.tokens.completion': usage.output_tokens,
                            'llm.tokens.total': usage.input_tokens
                            + usage.output_tokens,
                        },
                    )

            # Check if there's a tool use in the response
            for content_block in response.content:
                if content_block.type == 'tool_use':
                    return {
                        'content': response.content[0].text if response.content else '',
                        'function_call': {
                            'name': content_block.name,
                            'arguments': json.dumps(content_block.input),
                        },
                    }

            # Handle regular text response
            return {'content': response.content[0].text if response.content else ''}

        except Exception as e:
            raise Exception(f'Error in Claude API call: {str(e)}')

    @trace_llm_stream(provider='anthropic')
    async def stream(
        self,
        messages: List[Dict[str, str]],
        functions: Optional[List[Dict[str, Any]]] = None,
        **kwargs,
    ) -> AsyncIterator[Dict[str, Any]]:
        """Stream partial responses from the LLM as they are generated"""
        # Convert messages to Claude format
        system_message = next(
            (msg['content'] for msg in messages if msg['role'] == 'system'), None
        )

        conversation = []
        for msg in messages:
            if msg['role'] != 'system':
                conversation.append(
                    {
                        'role': 'assistant' if msg['role'] == 'assistant' else 'user',
                        'content': msg['content'],
                    }
                )

        anthropic_kwargs = {
            'model': self.model,
            'messages': conversation,
            'temperature': self.temperature,
            'max_tokens': self.kwargs.get('max_tokens', 1024),
            **self.kwargs,
            **kwargs,
        }

        if system_message:
            anthropic_kwargs['system'] = system_message

        if functions:
            anthropic_kwargs['tools'] = functions
        # Use Anthropic SDK streaming API and yield text deltas
        async with self.client.messages.stream(**anthropic_kwargs) as stream:
            async for event in stream:
                if (
                    getattr(event, 'type', None) == 'content_block_delta'
                    and hasattr(event, 'delta')
                    and getattr(event.delta, 'type', None) == 'text_delta'
                    and hasattr(event.delta, 'text')
                ):
                    yield {'content': event.delta.text}

    def get_message_content(self, response: Any) -> str:
        """Extract message content from response"""
        if isinstance(response, dict):
            return response.get('content', '')
        return str(response)

    def format_tool_for_llm(self, tool: 'Tool') -> Dict[str, Any]:
        """Format a single tool for Claude's API"""
        return {
            'type': 'custom',
            'name': tool.name,
            'description': tool.description,
            'input_schema': {
                'type': 'object',
                'properties': {
                    name: {
                        'type': info.get('type', 'string'),
                        'description': info.get('description', ''),
                        **(
                            {'items': info['items']}
                            if info.get('type') == 'array' and 'items' in info
                            else {}
                        ),
                    }
                    for name, info in tool.parameters.items()
                },
                'required': [
                    name
                    for name, info in tool.parameters.items()
                    if info.get('required', True)
                ],
            },
        }

    def format_tools_for_llm(self, tools: List['Tool']) -> List[Dict[str, Any]]:
        """Format tools for Claude's API"""
        return [self.format_tool_for_llm(tool) for tool in tools]

    def format_image_in_message(self, image: ImageMessage) -> str:
        """Format a image in the message"""
        raise NotImplementedError('Not implemented image for LLM Anthropic')



from .base_tool import Tool, ToolExecutionError
from .flo_tool import flo_tool, create_tool_from_function
from .partial_tool import PartialTool, create_partial_tool
from .tool_config import ToolConfig, create_tool_config

__all__ = [
    'Tool',
    'ToolExecutionError',
    'flo_tool',
    'create_tool_from_function',
    'PartialTool',
    'create_partial_tool',
    'ToolConfig',
    'create_tool_config',
]



from typing import Dict, Any, Optional
from .base_tool import Tool


class ToolConfig:
    """
    Represents a tool with optional pre-filled parameters.
    This allows the AgentBuilder to handle both regular tools and tools with configurations
    transparently through the same interface.
    """

    def __init__(
        self,
        tool: Tool,
        pre_filled_params: Optional[Dict[str, Any]] = None,
        name_override: Optional[str] = None,
        description_override: Optional[str] = None,
    ):
        """
        Create a tool configuration.

        Args:
            tool: The base tool
            pre_filled_params: Optional pre-filled parameters
            name_override: Optional custom name
            description_override: Optional custom description
        """
        self.tool = tool
        self.pre_filled_params = pre_filled_params or {}
        self.name_override = name_override
        self.description_override = description_override

    def is_partial(self) -> bool:
        """Check if this tool configuration has pre-filled parameters."""
        return bool(self.pre_filled_params)

    def to_tool(self) -> Tool:
        """Convert this configuration to a Tool (either original or partial)."""
        # If there are pre-filled parameters or custom name/description, create partial tool
        if (
            self.pre_filled_params
            or self.name_override is not None
            or self.description_override is not None
        ):
            # Import here to avoid circular imports
            from .partial_tool import PartialTool

            return PartialTool(
                base_tool=self.tool,
                pre_filled_params=self.pre_filled_params,
                name_override=self.name_override,
                description_override=self.description_override,
            )
        else:
            # No customizations, return original tool
            return self.tool


def create_tool_config(tool: Tool, **pre_filled_params) -> ToolConfig:
    """
    Create a tool configuration with pre-filled parameters.

    Args:
        tool: The base tool
        **pre_filled_params: Pre-filled parameters

    Returns:
        ToolConfig: A tool configuration
    """
    return ToolConfig(tool, pre_filled_params)



import inspect
from functools import partial
from flo_ai.arium.nodes import AriumNode, ForEachNode
from flo_ai.arium.protocols import ExecutableNode
from flo_ai.models.agent import Agent
from flo_ai.tool.base_tool import Tool
from flo_ai.utils.logger import logger
from typing import List, Optional, Callable, Literal, get_origin, get_args, Dict
from flo_ai.arium.models import StartNode, EndNode, Edge, default_router
from pathlib import Path


class BaseArium:
    def __init__(self):
        self.start_node_name = '__start__'
        self.end_node_names: set = set()  # Support multiple end nodes
        self.nodes: Dict[str, ExecutableNode | StartNode | EndNode] = dict()
        self.edges: Dict[str, Edge] = dict()

    def add_nodes(self, agents: List[ExecutableNode | StartNode | EndNode]):
        self.nodes.update({agent.name: agent for agent in agents})

    def start_at(self, node: ExecutableNode):
        start_node = StartNode()
        if start_node.name in self.nodes:
            raise ValueError(f'Start node {start_node.name} already exists')
        self.nodes[start_node.name] = start_node
        self.edges[start_node.name] = Edge(
            router_fn=partial(default_router, to_node=node.name), to_nodes=[node.name]
        )

    def add_end_to(self, node: ExecutableNode):
        # Create a unique end node name for this specific node
        end_node_name = f'__end__{node.name}__'
        end_node = EndNode()
        end_node.name = end_node_name

        # Add this end node name to our set of possible end nodes
        self.end_node_names.add(end_node_name)

        # Only add the end node if it doesn't exist yet
        if end_node_name not in self.nodes:
            self.nodes[end_node_name] = end_node

        self.edges[node.name] = Edge(
            router_fn=partial(default_router, to_node=end_node_name),
            to_nodes=[end_node_name],
        )

    def _check_router_return_type(self, router: Callable) -> Optional[List]:
        try:
            # Get the function signature
            sig = inspect.signature(router)
            return_annotation = sig.return_annotation

            # Check if there's no return annotation
            if return_annotation == inspect.Signature.empty:
                return None

            # Check if the return type is a Literal
            origin = get_origin(return_annotation)

            # In Python 3.8+, Literal types have get_origin() return typing.Literal
            if origin is Literal:
                # Extract the literal values
                literal_values = list(get_args(return_annotation))
                return literal_values

            return None

        except Exception as e:
            logger.error(f'Error checking router return type: {e}')
            return None

    def add_edge(
        self,
        from_node: str,
        to_nodes: List[str] = None,
        router: Optional[Callable] = None,
    ):
        if router and not callable(router):
            raise ValueError('Router must be a callable')

        if not to_nodes:
            raise ValueError('To nodes must be provided')

        if not router and len(to_nodes) != 1:
            raise ValueError(
                'Exactly one to node must be provided if router is not provided'
            )

        if from_node not in self.nodes:
            raise ValueError(f'Node {from_node} not found')

        wrong_nodes = [
            wrong_to_node
            for wrong_to_node in to_nodes
            if wrong_to_node not in self.nodes
        ]
        if wrong_nodes:
            raise ValueError(f'Nodes {wrong_nodes} not found')

        if router:
            literal_values = self._check_router_return_type(router)
            if literal_values is None:
                raise ValueError('Router return type is not a Literal')

            # Check if router supports self-reference
            supports_self_ref = getattr(router, 'supports_self_reference', False)

            # For self-referencing routers, we need to include the from_node in valid targets
            valid_targets = to_nodes.copy()
            if supports_self_ref:
                valid_targets.append(from_node)

            invalid_literals = [
                val for val in literal_values if val not in valid_targets
            ]
            if invalid_literals:
                raise ValueError(
                    f'Router return type includes literal values {invalid_literals} that are not in valid targets {valid_targets}'
                )

            # For self-referencing routers, allow router options to include from_node
            if supports_self_ref:
                # Router can return any of: to_nodes + from_node, but must include all to_nodes
                missing_targets = [
                    node for node in to_nodes if node not in literal_values
                ]
                if missing_targets:
                    raise ValueError(
                        f'Self-referencing router must include all to_nodes {to_nodes}, missing: {missing_targets}'
                    )
            else:
                # Non-self-referencing routers must match exactly
                if set(literal_values) != set(to_nodes):
                    raise ValueError(
                        f'Router return type values {literal_values} do not match to_nodes {to_nodes}'
                    )

        self.edges[from_node] = Edge(
            router_fn=router
            if router
            else partial(default_router, to_node=to_nodes[0]),
            to_nodes=to_nodes,
        )

    def check_orphan_nodes(self) -> List[str]:
        if not self.nodes:
            return []

        # Get all nodes with outgoing edges
        nodes_with_outgoing = set(self.edges.keys())

        # Get all nodes with incoming edges by examining router return types and to_nodes
        nodes_with_incoming = set()

        for _, target in self.edges.items():
            nodes_with_incoming.update(target.to_nodes)

        # Find orphan nodes: nodes that have neither incoming nor outgoing edges
        all_nodes = set(self.nodes.keys())
        connected_nodes = nodes_with_outgoing.union(nodes_with_incoming)
        orphan_nodes = all_nodes - connected_nodes

        return list(orphan_nodes)

    def validate_graph(self) -> bool:
        orphan_nodes = self.check_orphan_nodes()

        if orphan_nodes:
            raise ValueError(
                f'Orphan nodes found: {orphan_nodes}. These nodes have no incoming or outgoing edges.'
            )

        # Check for exactly 1 start node
        start_nodes = [
            node for node in self.nodes.values() if isinstance(node, StartNode)
        ]
        if len(start_nodes) == 0:
            raise ValueError(
                f'Graph must have exactly 1 start node. Found 0 start nodes: {start_nodes}'
            )
        elif len(start_nodes) > 1:
            raise ValueError(
                f'Graph must have exactly 1 start node. Found {len(start_nodes)} start nodes.'
            )

        # Check for at least 1 end node
        end_nodes = [node for node in self.nodes.values() if isinstance(node, EndNode)]
        if len(end_nodes) == 0:
            raise ValueError('Graph must have at least 1 end node. Found 0 end nodes.')

        return True

    def visualize_graph(
        self,
        output_path: str = 'graph_visualization.png',
        graph_title: str = 'Arium Graph Visualization',
        figsize: tuple = (12, 8),
        node_size: int = 3000,
        font_size: int = 10,
        dpi: int = 300,
    ) -> None:
        """
        Generate a graph visualization and save it as PNG.

        Args:
            output_path: Path where the PNG file will be saved
            figsize: Figure size as (width, height) in inches
            node_size: Size of the nodes in the graph
            font_size: Font size for node labels
            dpi: Resolution of the saved image
        """
        import matplotlib.pyplot as plt
        import matplotlib.patches as patches
        import networkx as nx

        if not self.nodes:
            logger.error('No nodes to visualize')
            return

        # Create directed graph
        G = nx.DiGraph()

        # Add nodes with their types
        for node_name, node in self.nodes.items():
            node_type = self._get_node_type(node)
            G.add_node(node_name, node_type=node_type, node_obj=node)

        # Add edges
        for from_node, edge in self.edges.items():
            for to_node in edge.to_nodes:
                if to_node in self.nodes:
                    G.add_edge(from_node, to_node, edge_obj=edge)

        # Create matplotlib figure
        fig, ax = plt.subplots(figsize=figsize, dpi=dpi)

        # Use hierarchical layout for better DAG visualization
        try:
            # Try different layouts for better DAG appearance
            pos = nx.nx_agraph.graphviz_layout(G, prog='dot', args='-Grankdir=TB')
        except Exception as e:
            logger.error(f'Error in graphviz_layout: {e}')
            try:
                # Fallback to planar layout for DAG structure
                pos = nx.planar_layout(G)
            except Exception as e:
                logger.error(f'Error in graphviz_layout: {e}')
                # Final fallback to shell layout for better hierarchy
                pos = nx.shell_layout(G)

        # Define colors for different node types
        node_colors = {
            'start': '#90EE90',  # Light green
            'end': '#FFB6C1',  # Light pink
            'agent': '#87CEEB',  # Sky blue
            'tool': '#DDA0DD',  # Plum
        }

        # Draw nodes with different colors based on type
        for node_name, node_data in G.nodes(data=True):
            node_type = node_data['node_type']
            color = node_colors.get(node_type, '#CCCCCC')

            nx.draw_networkx_nodes(
                G,
                pos,
                nodelist=[node_name],
                node_color=color,
                node_size=node_size,
                alpha=0.9,
                ax=ax,
            )

        # Separate edges by router type for different styling
        default_edges = []
        custom_edges = []

        for edge_data in G.edges(data=True):
            from_node, to_node, data = edge_data
            edge_obj = data['edge_obj']

            if edge_obj.is_default_router():
                default_edges.append((from_node, to_node))
            else:
                custom_edges.append((from_node, to_node))

        # Draw default router edges with dotted lines
        if default_edges:
            nx.draw_networkx_edges(
                G,
                pos,
                edgelist=default_edges,
                edge_color='gray',
                arrows=True,
                arrowsize=20,
                arrowstyle='->',
                connectionstyle='arc3,rad=0.1',
                style='solid',
                width=2,
                ax=ax,
            )

        # Draw custom router edges with solid lines
        if custom_edges:
            nx.draw_networkx_edges(
                G,
                pos,
                edgelist=custom_edges,
                edge_color='black',
                arrows=True,
                arrowsize=20,
                arrowstyle='->',
                connectionstyle='arc3,rad=0.1',
                style='dotted',
                width=2,
                ax=ax,
            )

        # Add labels
        nx.draw_networkx_labels(G, pos, font_size=font_size, font_weight='bold', ax=ax)

        # Add title
        plt.title(graph_title, fontsize=16, fontweight='bold', pad=20)

        # Add legend
        legend_elements = [
            patches.Patch(color='#90EE90', label='Start Node'),
            patches.Patch(color='#FFB6C1', label='End Node'),
            patches.Patch(color='#87CEEB', label='Agent'),
            patches.Patch(color='#DDA0DD', label='Tool'),
            patches.Patch(
                facecolor='none',
                edgecolor='gray',
                linestyle='dotted',
                linewidth=2,
                label='Default Router',
            ),
            patches.Patch(
                facecolor='none',
                edgecolor='black',
                linestyle='solid',
                linewidth=2,
                label='Custom Router',
            ),
        ]
        ax.legend(handles=legend_elements, loc='upper right', bbox_to_anchor=(1.15, 1))

        # Remove axes
        ax.axis('off')

        # Adjust layout to prevent legend cutoff
        plt.tight_layout()

        # Save the figure
        output_path = Path(output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)

        plt.savefig(
            str(output_path),
            format='png',
            dpi=dpi,
            bbox_inches='tight',
            facecolor='white',
            edgecolor='none',
        )
        plt.close()

        logger.info(f'Graph visualization saved to: {output_path}')

    def _get_node_type(self, node) -> str:
        """Helper method to determine node type for visualization."""
        if isinstance(node, StartNode):
            return 'start'
        elif isinstance(node, EndNode):
            return 'end'
        elif isinstance(node, Agent):
            return 'agent'
        elif isinstance(node, Tool):
            return 'tool'
        elif isinstance(node, ForEachNode):
            return 'foreach'
        elif isinstance(node, AriumNode):
            return 'arium'
        else:
            return 'unknown'



from typing import Dict, Any, List, Optional, AsyncIterator
import aiohttp
import json
from .base_llm import BaseLLM, ImageMessage
from flo_ai.tool.base_tool import Tool
from flo_ai.telemetry.instrumentation import trace_llm_call, trace_llm_stream


class OllamaLLM(BaseLLM):
    def __init__(
        self,
        model: str = 'llama2',
        api_key: str = None,
        temperature: float = 0.7,
        base_url: str = 'http://localhost:11434',
        **kwargs,
    ):
        super().__init__(model, api_key, temperature, **kwargs)
        self.base_url = base_url.rstrip('/')

    @trace_llm_call(provider='ollama')
    async def generate(
        self,
        messages: List[Dict[str, str]],
        functions: Optional[List[Dict[str, Any]]] = None,
        output_schema: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        # Convert messages to Ollama format
        prompt = ''
        for msg in messages:
            role = msg['role']
            content = msg['content']
            if role == 'system':
                prompt += f'System: {content}\n'
            elif role == 'user':
                prompt += f'User: {content}\n'
            elif role == 'assistant':
                prompt += f'Assistant: {content}\n'

        # Add output schema instruction if provided
        if output_schema:
            prompt += f'\nPlease provide your response in JSON format according to this schema:\n{json.dumps(output_schema, indent=2)}\n'

        # Prepare request payload
        payload = {
            'model': self.model,
            'prompt': prompt,
            'temperature': self.temperature,
            'stream': False,
            **self.kwargs,
        }

        # Add function information if provided
        if functions:
            payload['functions'] = functions

        async with aiohttp.ClientSession() as session:
            async with session.post(
                f'{self.base_url}/api/generate', json=payload
            ) as response:
                if response.status != 200:
                    raise Exception(f'Ollama API error: {await response.text()}')

                result = await response.json()
                return {
                    'content': result.get('response', ''),
                    'function_call': result.get('function_call'),
                }

    @trace_llm_stream(provider='ollama')
    async def stream(
        self,
        messages: List[Dict[str, str]],
        functions: Optional[List[Dict[str, Any]]] = None,
    ) -> AsyncIterator[Dict[str, Any]]:
        """Stream partial responses from the hosted Ollama service.

        Note: For streaming, do not include the 'stream' flag in payload; the
        service defaults to streamed output.
        """
        # Convert messages to Ollama prompt format
        prompt = ''
        for msg in messages:
            role = msg['role']
            content = msg['content']
            if role == 'system':
                prompt += f'System: {content}\n'
            elif role == 'user':
                prompt += f'User: {content}\n'
            elif role == 'assistant':
                prompt += f'Assistant: {content}\n'

        # Prepare request payload without 'stream' key for streaming
        payload = {
            'model': self.model,
            'prompt': prompt,
            'temperature': self.temperature,
            **self.kwargs,
        }

        if functions:
            payload['functions'] = functions

        async with aiohttp.ClientSession() as session:
            async with session.post(
                f'{self.base_url}/api/generate', json=payload
            ) as response:
                if response.status != 200:
                    raise Exception(f'Ollama API error: {await response.text()}')

                async for raw_line in response.content:
                    line = raw_line.decode('utf-8').strip()
                    if not line:
                        continue
                    try:
                        data = json.loads(line)
                    except Exception:
                        # Skip non-JSON lines
                        continue

                    if 'response' in data and data['response']:
                        yield {'content': data['response']}

                    if data.get('done') is True:
                        break

    def get_message_content(self, response: Any) -> str:
        """Extract message content from response"""
        if isinstance(response, dict):
            return response.get('content', '')
        return str(response)

    def format_tool_for_llm(self, tool: 'Tool') -> Dict[str, Any]:
        """Format a single tool for Ollama's API"""
        return {
            'name': tool.name,
            'description': tool.description,
            'parameters': {
                'type': 'object',
                'properties': {
                    name: {
                        'type': info.get('type', 'string'),
                        'description': info.get('description', ''),
                        **(
                            {'items': info['items']}
                            if info.get('type') == 'array' and 'items' in info
                            else {}
                        ),
                    }
                    for name, info in tool.parameters.items()
                },
                'required': [
                    name
                    for name, info in tool.parameters.items()
                    if info.get('required', True)
                ],
            },
        }

    def format_tools_for_llm(self, tools: List['Tool']) -> List[Dict[str, Any]]:
        """Format tools for Ollama's API"""
        return [self.format_tool_for_llm(tool) for tool in tools]

    def format_image_in_message(self, image: ImageMessage) -> str:
        """Format a image in the message"""
        raise NotImplementedError('Not implemented image for LLM Ollama')



from enum import Enum
from typing import AsyncIterator, Dict, Any, List, Optional
from datetime import datetime, timedelta
import jwt
from .base_llm import BaseLLM, ImageMessage
from .openai_llm import OpenAI
from .gemini_llm import Gemini
from .anthropic_llm import Anthropic
from flo_ai.tool.base_tool import Tool


class LLMProvider(Enum):
    """Enum for supported LLM providers"""

    OPENAI = 'openai'
    GEMINI = 'gemini'
    ANTHROPIC = 'anthropic'


class RootFloLLM(BaseLLM):
    """
    Proxy LLM class that routes to different SDK implementations based on type.
    Acts as a unified interface to OpenAI, Gemini, and Anthropic SDKs via a proxy URL.
    """

    def __init__(
        self,
        base_url: str,
        model_id: str,
        llm_model: str,
        llm_provider: LLMProvider,
        app_key: str,
        app_secret: str,
        issuer: str,
        audience: str,
        temperature: float = 0.7,
        **kwargs,
    ):
        """
        Initialize RootFloLLM proxy.

        Args:
            base_url: The base URL of the proxy server
            model_id: The model identifier
            llm_provider: Type of LLM SDK to use (LLMProvider enum)
            app_key: Application key for X-Rootflo-Key header
            app_secret: Application secret for JWT signing
            issuer: JWT issuer claim
            audience: JWT audience claim
            temperature: Temperature parameter for generation
            **kwargs: Additional parameters to pass to the underlying SDK
        """
        # Generate JWT token
        now = datetime.now()
        payload = {
            'iss': issuer,
            'aud': audience,
            'iat': int(now.timestamp()),
            'exp': int((now + timedelta(seconds=3600)).timestamp()),
            'role_id': 'floconsole-service',
            'user_id': 'service',
            'service_auth': True,
        }
        service_token = jwt.encode(payload, app_secret, algorithm='HS256')
        api_token = f'fc_{service_token}'

        super().__init__(
            model=llm_model, api_key=api_token, temperature=temperature, **kwargs
        )

        self.base_url = base_url
        self.model_id = model_id
        self.llm_provider = llm_provider

        # Construct full URL
        full_url = f'{base_url}/{model_id}'

        # Prepare custom headers for proxy authentication
        custom_headers = {'X-Rootflo-Key': app_key}

        # Instantiate appropriate SDK wrapper based on llm_provider
        if llm_provider == LLMProvider.OPENAI:
            self._llm = OpenAI(
                model=llm_model,
                base_url=full_url,
                api_key=api_token,
                temperature=temperature,
                custom_headers=custom_headers,
                **kwargs,
            )
        elif llm_provider == LLMProvider.ANTHROPIC:
            self._llm = Anthropic(
                model=llm_model,
                base_url=full_url,
                api_key=api_token,
                temperature=temperature,
                custom_headers=custom_headers,
                **kwargs,
            )
        elif llm_provider == LLMProvider.GEMINI:
            # Gemini SDK - pass base_url which will be handled via http_options
            self._llm = Gemini(
                model=llm_model,
                api_key=api_token,
                temperature=temperature,
                base_url=full_url,
                custom_headers=custom_headers,
                **kwargs,
            )
        else:
            raise ValueError(f'Unsupported LLM provider: {llm_provider}')

    async def generate(
        self,
        messages: List[Dict[str, str]],
        functions: Optional[List[Dict[str, Any]]] = None,
        output_schema: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """Generate a response from the LLM"""
        return await self._llm.generate(
            messages, functions=functions, output_schema=output_schema, **kwargs
        )

    async def stream(
        self,
        messages: List[Dict[str, Any]],
        functions: Optional[List[Dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> AsyncIterator[Dict[str, Any]]:
        """Generate a streaming response from the LLM"""
        async for chunk in self._llm.stream(messages, functions=functions, **kwargs):
            yield chunk

    def get_message_content(self, response: Any) -> str:
        """Extract message content from response"""
        return self._llm.get_message_content(response)

    def format_tool_for_llm(self, tool: 'Tool') -> Dict[str, Any]:
        """Format a tool for the specific LLM's API"""
        return self._llm.format_tool_for_llm(tool)

    def format_tools_for_llm(self, tools: List['Tool']) -> List[Dict[str, Any]]:
        """Format a list of tools for the specific LLM's API"""
        return self._llm.format_tools_for_llm(tools)

    def format_image_in_message(self, image: ImageMessage) -> str:
        """Format a image in the message"""
        return self._llm.format_image_in_message(image)



from typing import List, Optional, Dict, Any, Union, Type
import yaml
from flo_ai.models.agent import Agent
from flo_ai.models.base_agent import ReasoningPattern
from flo_ai.llm import BaseLLM, OpenAI, Anthropic, Gemini, OllamaLLM, VertexAI
from flo_ai.tool.base_tool import Tool
from flo_ai.tool.tool_config import ToolConfig, create_tool_config
from flo_ai.formatter.yaml_format_parser import FloYamlParser
from pydantic import BaseModel


class AgentBuilder:
    """
    A facade class that simplifies the creation and configuration of AI agents.
    """

    def __init__(self):
        self._name = 'AI Assistant'
        self._system_prompt = 'You are a helpful AI assistant.'
        self._llm: Optional[BaseLLM] = None
        self._tools: List[Tool] = []
        self._max_retries = 3
        self._reasoning_pattern = ReasoningPattern.DIRECT
        self._output_schema: Optional[Dict[str, Any]] = None
        self._role: Optional[str] = None
        self._act_as: Optional[str] = None

    def with_name(self, name: str) -> 'AgentBuilder':
        """Set the agent's name"""
        self._name = name
        return self

    def with_prompt(self, system_prompt: str) -> 'AgentBuilder':
        """Set the system prompt"""
        self._system_prompt = system_prompt
        return self

    def with_llm(self, llm: BaseLLM) -> 'AgentBuilder':
        """Configure the LLM to use

        Args:
            llm: An instance of a BaseLLM implementation
        """
        self._llm = llm
        return self

    def with_tools(
        self, tools: Union[List[Tool], List[ToolConfig], List[Dict[str, Any]]]
    ) -> 'AgentBuilder':
        """
        Add tools to the agent.

        Args:
            tools: List of tools, tool configurations, or tool dictionaries.
                   Each tool dictionary should have:
                   - 'tool': The Tool object
                   - 'pre_filled_params': Optional dict of pre-filled parameters
                   - 'name_override': Optional custom name
                   - 'description_override': Optional custom description

        Examples:
            # Regular tools
            builder.with_tools([tool1, tool2])

            # Tool configurations
            builder.with_tools([
                ToolConfig(tool1, pre_filled_params={"param1": "value1"}),
                ToolConfig(tool2, pre_filled_params={"param2": "value2"})
            ])

            # Tool dictionaries
            builder.with_tools([
                {"tool": tool1, "pre_filled_params": {"param1": "value1"}},
                {"tool": tool2, "pre_filled_params": {"param2": "value2"}}
            ])
        """
        processed_tools = []

        for tool_item in tools:
            if isinstance(tool_item, Tool):
                # Regular tool - add as is
                processed_tools.append(tool_item)
            elif isinstance(tool_item, ToolConfig):
                # Tool configuration - convert to tool
                processed_tools.append(tool_item.to_tool())
            elif isinstance(tool_item, dict):
                # Tool dictionary - convert to ToolConfig then to tool
                tool = tool_item['tool']
                pre_filled_params = tool_item.get('pre_filled_params', {})
                name_override = tool_item.get('name_override')
                description_override = tool_item.get('description_override')

                tool_config = ToolConfig(
                    tool=tool,
                    pre_filled_params=pre_filled_params,
                    name_override=name_override,
                    description_override=description_override,
                )
                processed_tools.append(tool_config.to_tool())
            else:
                raise ValueError(f'Unsupported tool type: {type(tool_item)}')

        self._tools = processed_tools
        return self

    def add_tool(self, tool: Tool, **pre_filled_params) -> 'AgentBuilder':
        """
        Add a single tool with optional pre-filled parameters.

        Args:
            tool: The tool to add
            **pre_filled_params: Pre-filled parameters for the tool

        Example:
            builder.add_tool(
                bigquery_tool,
                datasource_id="ds_123",
                project_id="my-project"
            )
        """
        if pre_filled_params:
            tool_config = create_tool_config(tool, **pre_filled_params)
            self._tools.append(tool_config.to_tool())
        else:
            self._tools.append(tool)
        return self

    def with_reasoning(self, pattern: ReasoningPattern) -> 'AgentBuilder':
        """Set the reasoning pattern"""
        self._reasoning_pattern = pattern
        return self

    def with_retries(self, max_retries: int) -> 'AgentBuilder':
        """Set maximum number of retries"""
        self._max_retries = max_retries
        return self

    def with_output_schema(
        self, schema: Union[Dict[str, Any], Type[BaseModel]]
    ) -> 'AgentBuilder':
        """Set output schema for structured responses

        Args:
            schema: Either a JSON schema dictionary or a Pydantic model class
        """
        if isinstance(schema, type) and issubclass(schema, BaseModel):
            self._output_schema = schema.model_json_schema()
        else:
            self._output_schema = schema
        return self

    def with_role(self, role: str) -> 'AgentBuilder':
        """Set the agent's role"""
        self._role = role
        return self

    def with_actas(self, act_as: str) -> 'AgentBuilder':
        """Set the agent's role"""
        self._act_as = act_as
        return self

    def build(self) -> Agent:
        """Build and return the configured agent"""
        if not self._llm:
            raise ValueError('LLM must be configured before building the agent')

        return Agent(
            name=self._name,
            system_prompt=self._system_prompt,
            llm=self._llm,
            tools=self._tools,
            max_retries=self._max_retries,
            reasoning_pattern=self._reasoning_pattern,
            output_schema=self._output_schema,
            role=self._role,
            act_as=self._act_as,
        )

    @classmethod
    def from_yaml(
        cls,
        yaml_str: str,
        tools: Optional[List[Tool]] = None,
        base_llm: Optional[BaseLLM] = None,
        tool_registry: Optional[Dict[str, Tool]] = None,
    ) -> 'AgentBuilder':
        """Create an agent builder from a YAML configuration string

        Args:
            yaml_str: YAML string containing agent configuration
            tools: Optional list of tools to use with the agent
            base_llm: Optional base LLM to use
            tool_registry: Optional dictionary mapping tool names to Tool objects
                          Used to resolve tool references in YAML

        Returns:
            AgentBuilder: Configured agent builder instance
        """
        config = yaml.safe_load(yaml_str)

        if 'agent' not in config:
            raise ValueError('YAML must contain an "agent" section')

        agent_config = config['agent']
        builder = cls()

        # Set basic properties
        builder.with_name(agent_config.get('name', 'AI Assistant'))
        builder.with_prompt(agent_config.get('job', 'You are a helpful AI assistant.'))
        builder.with_role(agent_config.get('role'))
        builder.with_actas(agent_config.get('act_as'))

        # Configure LLM based on model settings
        if 'model' in agent_config and base_llm is None:
            base_url = agent_config.get('base_url', None)
            model_config: dict = agent_config['model']
            provider = model_config.get('provider', 'openai').lower()
            model_name = model_config.get('name')

            if not model_name:
                raise ValueError('Model name must be specified in YAML configuration')

            if provider == 'openai':
                builder.with_llm(OpenAI(model=model_name, base_url=base_url))
            elif provider == 'anthropic':
                builder.with_llm(Anthropic(model=model_name, base_url=base_url))
            elif provider == 'gemini':
                builder.with_llm(Gemini(model=model_name, base_url=base_url))
            elif provider == 'ollama':
                builder.with_llm(OllamaLLM(model=model_name, base_url=base_url))
            elif provider == 'vertexai':
                project = model_config.get('project')
                location = model_config.get('location', 'asia-south1')
                builder.with_llm(
                    VertexAI(
                        model=model_name,
                        project=project,
                        location=location,
                        base_url=base_url,
                    )
                )
            else:
                raise ValueError(f'Unsupported model provider: {provider}')
        else:
            if base_llm is None:
                raise ValueError(
                    'Model must be specified in YAML configuration or base_llm must be provided'
                )
            builder.with_llm(base_llm)

        # Handle tools configuration
        if 'tools' in agent_config:
            # Process tools from YAML configuration
            yaml_tools = cls._process_yaml_tools(agent_config['tools'], tool_registry)
            builder.with_tools(yaml_tools)
        elif tools:
            # Use provided tools
            builder.with_tools(tools)

        # Set parser if present
        if 'parser' in agent_config:
            parser = FloYamlParser.create(yaml_dict=config)
            builder.with_output_schema(parser.get_format())

        # Apply settings if present
        if 'settings' in agent_config:
            settings = agent_config['settings']
            if 'temperature' in settings:
                builder._llm.temperature = settings['temperature']
            if 'max_retries' in settings:
                builder.with_retries(settings['max_retries'])
            if 'reasoning_pattern' in settings:
                builder.with_reasoning(ReasoningPattern[settings['reasoning_pattern']])

        return builder

    @classmethod
    def _process_yaml_tools(
        cls,
        tools_config: List[Dict[str, Any]],
        tool_registry: Optional[Dict[str, Tool]] = None,
    ) -> List[Tool]:
        """Process tools configuration from YAML.

        Args:
            tools_config: List of tool configurations from YAML
            tool_registry: Optional dictionary mapping tool names to Tool objects

        Returns:
            List[Tool]: Processed tools
        """
        processed_tools = []

        for tool_config in tools_config:
            if isinstance(tool_config, str):
                # Simple string reference - look up in registry
                if tool_registry and tool_config in tool_registry:
                    processed_tools.append(tool_registry[tool_config])
                else:
                    raise ValueError(f"Tool '{tool_config}' not found in tool registry")
            elif isinstance(tool_config, dict):
                # Tool configuration dictionary
                tool_name = tool_config.get('name')
                if not tool_name:
                    raise ValueError("Tool configuration must have a 'name' field")

                # Look up tool in registry
                if tool_registry and tool_name in tool_registry:
                    base_tool = tool_registry[tool_name]
                else:
                    raise ValueError(f"Tool '{tool_name}' not found in tool registry")

                # Extract configuration
                pre_filled_params = tool_config.get('pre_filled_params', {})
                name_override = tool_config.get('name_override')
                description_override = tool_config.get('description_override')

                # Create tool configuration
                tool_config_obj = ToolConfig(
                    tool=base_tool,
                    pre_filled_params=pre_filled_params,
                    name_override=name_override,
                    description_override=description_override,
                )

                # If there are pre-filled parameters or custom name/description, convert to tool
                if (
                    pre_filled_params
                    or name_override is not None
                    or description_override is not None
                ):
                    processed_tools.append(tool_config_obj.to_tool())
                else:
                    # No pre-filled params and no custom name/description, use original tool
                    processed_tools.append(base_tool)
            else:
                raise ValueError(
                    f'Invalid tool configuration type: {type(tool_config)}'
                )

        return processed_tools



"""
Plan Execution Agents for Flo AI Framework

This module provides standard agent classes for plan-based execution patterns,
making it easy to create plan-and-execute workflows.
"""

from typing import List, Optional
from flo_ai.models.agent import Agent
from flo_ai.llm.base_llm import BaseLLM
from flo_ai.arium.memory import PlanAwareMemory
from flo_ai.tool.plan_tool import PlanTool, StepTool, PlanStatusTool


class PlannerAgent(Agent):
    """
    Agent specialized for creating execution plans.

    Automatically equipped with tools to store plans in PlanAwareMemory.
    """

    def __init__(
        self,
        memory: PlanAwareMemory,
        llm: BaseLLM,
        name: str = 'planner',
        system_prompt: Optional[str] = None,
        **kwargs,
    ):
        """
        Initialize the planner agent.

        Args:
            memory: PlanAwareMemory instance to store plans in
            llm: LLM instance for the agent
            name: Agent name (default: "planner")
            system_prompt: Custom system prompt, or uses default if None
            **kwargs: Additional arguments for Agent
        """

        # Default system prompt for planners
        if system_prompt is None:
            system_prompt = """You are a project planner. Create execution plans in this EXACT format:

EXECUTION PLAN: [Clear Title]
DESCRIPTION: [Brief description]

STEPS:
1. step_1: [Description] → [agent_name]
2. step_2: [Description] → [agent_name] (depends on: step_1)
3. step_3: [Description] → [agent_name] (depends on: step_1, step_2)

Rules:
- Use clear, actionable step descriptions
- Assign steps to appropriate agent names
- Include dependencies where steps must be done in sequence
- Keep step IDs simple (step_1, step_2, etc.)

IMPORTANT: After generating the plan text, use the store_execution_plan tool to save it."""

        # Create plan tool
        plan_tool = PlanTool(memory)
        plan_status_tool = PlanStatusTool(memory)

        super().__init__(
            name=name,
            system_prompt=system_prompt,
            llm=llm,
            tools=[plan_tool, plan_status_tool],
            **kwargs,
        )


class ExecutorAgent(Agent):
    """
    Agent specialized for executing plan steps.

    Automatically equipped with tools to mark steps as completed in PlanAwareMemory.
    """

    def __init__(
        self,
        memory: PlanAwareMemory,
        llm: BaseLLM,
        name: str,
        system_prompt: Optional[str] = None,
        **kwargs,
    ):
        """
        Initialize the executor agent.

        Args:
            memory: PlanAwareMemory instance to update plans in
            llm: LLM instance for the agent
            name: Agent name (must match agent names used in plans)
            system_prompt: Custom system prompt, or uses default if None
            **kwargs: Additional arguments for Agent
        """

        # Default system prompt for executors
        if system_prompt is None:
            system_prompt = f"""You are a {name} executing specific steps from an execution plan.

Process:
1. Check the current execution plan status using check_plan_status
2. Look for steps assigned to "{name}" that are ready to execute
3. Execute the step and provide detailed results
4. Use complete_step tool to mark the step as completed with your results

Focus on providing high-quality, detailed work for each step you execute."""

        # Create step tools
        step_tool = StepTool(memory, name)
        plan_status_tool = PlanStatusTool(memory)

        super().__init__(
            name=name,
            system_prompt=system_prompt,
            llm=llm,
            tools=[step_tool, plan_status_tool],
            **kwargs,
        )


def create_plan_execution_agents(
    memory: PlanAwareMemory,
    llm: BaseLLM,
    executor_agents: List[str],
    planner_name: str = 'planner',
) -> dict:
    """
    Factory function to create a complete set of plan execution agents.

    Args:
        memory: PlanAwareMemory instance
        llm: LLM instance for all agents
        executor_agents: List of executor agent names (e.g., ["developer", "tester", "reviewer"])
        planner_name: Name for the planner agent

    Returns:
        Dict mapping agent names to agent instances
    """
    agents = {}

    # Create planner
    agents[planner_name] = PlannerAgent(memory=memory, llm=llm, name=planner_name)

    # Create executors
    for agent_name in executor_agents:
        agents[agent_name] = ExecutorAgent(memory=memory, llm=llm, name=agent_name)

    return agents


def create_software_development_agents(memory: PlanAwareMemory, llm: BaseLLM) -> dict:
    """
    Create a standard set of agents for software development workflows.

    Args:
        memory: PlanAwareMemory instance
        llm: LLM instance for all agents

    Returns:
        Dict with planner, developer, tester, and reviewer agents
    """
    agents = {}

    # Planner with software development focus
    agents['planner'] = PlannerAgent(
        memory=memory,
        llm=llm,
        name='planner',
        system_prompt="""You are a software development project planner. Create execution plans for development tasks.

EXECUTION PLAN: [Clear Title]
DESCRIPTION: [Brief description]

STEPS:
1. step_1: [Development task] → developer
2. step_2: [Development task] → developer (depends on: step_1)
3. step_3: [Testing task] → tester (depends on: step_2)
4. step_4: [Review task] → reviewer (depends on: step_3)

Use these agents: developer, tester, reviewer
Focus on breaking down development tasks into logical, sequential steps.

IMPORTANT: After generating the plan, use store_execution_plan to save it.""",
    )

    # Developer
    agents['developer'] = ExecutorAgent(
        memory=memory,
        llm=llm,
        name='developer',
        system_prompt="""You are a software developer executing implementation steps.
Provide detailed code implementations, technical designs, and development work.
Always check the plan status first, then execute your assigned steps thoroughly.""",
    )

    # Tester
    agents['tester'] = ExecutorAgent(
        memory=memory,
        llm=llm,
        name='tester',
        system_prompt="""You are a QA tester validating implementations.
Create comprehensive test scenarios, validate functionality, and report results.
Always check the plan status first, then execute your assigned testing steps.""",
    )

    # Reviewer
    agents['reviewer'] = ExecutorAgent(
        memory=memory,
        llm=llm,
        name='reviewer',
        system_prompt="""You are a code reviewer providing final validation.
Review completed work, check quality, and provide final approval or feedback.
Always check the plan status first, then execute your assigned review steps.""",
    )

    return agents



"""
Document-related data models for Flo AI framework.

This module contains document types and message classes to avoid circular imports.
"""

from dataclasses import dataclass
from enum import Enum
from typing import Dict, Any, Optional


class DocumentType(Enum):
    """Enumeration of supported document types."""

    PDF = 'application/pdf'
    TXT = 'text/plain'


@dataclass
class DocumentMessage:
    """
    Data structure for document inputs to LLMs.

    Supports various document formats with extensible design for future types.
    """

    document_type: DocumentType
    document_url: Optional[str] = None
    document_bytes: Optional[bytes] = None
    document_file_path: Optional[str] = None
    document_base64: Optional[str] = None
    mime_type: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = None



"""
Models package for flo_ai - Agent framework components
"""

from .agent import Agent, MessageType
from .agent_error import AgentError
from .base_agent import BaseAgent, AgentType, ReasoningPattern
from .document import DocumentMessage, DocumentType
from .chat_message import ChatMessage

__all__ = [
    'Agent',
    'AgentError',
    'BaseAgent',
    'AgentType',
    'ReasoningPattern',
    'DocumentMessage',
    'DocumentType',
    'MessageType',
    'ChatMessage',
]



"""
LLM-Powered Router Functions for Arium Workflows

This module provides intelligent routing capabilities using Large Language Models
to make dynamic routing decisions based on conversation context and history.
"""

from abc import ABC, abstractmethod
from typing import Dict, Optional, Callable, Any, Union, get_args, List
from functools import wraps
from flo_ai.arium.memory import BaseMemory, ExecutionPlan, StepStatus
from flo_ai.llm.base_llm import BaseLLM
from flo_ai.llm import OpenAI
from flo_ai.utils.logger import logger


class BaseLLMRouter(ABC):
    """
    Base class for LLM-powered routers that make intelligent routing decisions
    based on conversation context and history.
    """

    def __init__(
        self,
        llm: Optional[BaseLLM] = None,
        temperature: float = 0.1,
        max_retries: int = 3,
        fallback_strategy: str = 'first',
    ):
        """
        Initialize the LLM router.

        Args:
            llm: The LLM instance to use for routing decisions. Defaults to GPT-4o-mini.
            temperature: Temperature for LLM calls (lower = more deterministic)
            max_retries: Maximum number of retries for LLM calls
            fallback_strategy: Strategy when LLM fails ("first", "last", "random")
        """
        self.llm = llm or OpenAI(model='gpt-4o-mini', temperature=temperature)
        self.temperature = temperature
        self.max_retries = max_retries
        self.fallback_strategy = fallback_strategy
        self.supports_self_reference = (
            False  # Most routers don't support self-reference by default
        )

    @abstractmethod
    def get_routing_options(self) -> Dict[str, str]:
        """
        Return a dictionary mapping route names to their descriptions.

        Returns:
            Dict[str, str]: Mapping of route names to descriptions
        """
        pass

    @abstractmethod
    def get_routing_prompt(
        self,
        memory: BaseMemory,
        options: Dict[str, str],
        execution_context: dict = None,
    ) -> str:
        """
        Generate the prompt for the LLM to make routing decisions.

        Args:
            memory: The conversation memory
            options: Available routing options with descriptions

        Returns:
            str: The prompt for the LLM
        """
        pass

    def get_fallback_route(self, options: Dict[str, str]) -> str:
        """
        Get fallback route when LLM fails.

        Args:
            options: Available routing options

        Returns:
            str: The fallback route name
        """
        routes = list(options.keys())

        if self.fallback_strategy == 'first':
            return routes[0]
        elif self.fallback_strategy == 'last':
            return routes[-1]
        elif self.fallback_strategy == 'random':
            import random

            return random.choice(routes)
        else:
            return routes[0]

    async def route(self, memory: BaseMemory, execution_context: dict = None) -> str:
        """
        Make a routing decision using the LLM.

        Args:
            memory: The conversation memory

        Returns:
            str: The name of the route to take
        """
        options = self.get_routing_options()

        for attempt in range(self.max_retries):
            try:
                prompt = self.get_routing_prompt(memory, options, execution_context)

                messages = [{'role': 'user', 'content': prompt}]
                response = await self.llm.generate(messages)
                decision = self.llm.get_message_content(response).strip().lower()

                # Find matching option (case-insensitive)
                for option_name in options:
                    if (
                        option_name.lower() == decision
                        or option_name.lower() in decision
                    ):
                        logger.info(f'LLM router selected: {option_name}')
                        return option_name

                # If no exact match, try partial matching
                for option_name in options:
                    if (
                        decision in option_name.lower()
                        or option_name.lower() in decision
                    ):
                        logger.info(
                            f'LLM router selected (partial match): {option_name}'
                        )
                        return option_name

                logger.warning(
                    f"LLM router attempt {attempt + 1}: Invalid decision '{decision}', retrying..."
                )

            except Exception as e:
                logger.error(f'LLM router attempt {attempt + 1} failed: {e}')

        # Fallback strategy
        fallback = self.get_fallback_route(options)
        logger.warning(f'LLM router failed, using fallback: {fallback}')
        return fallback


class SmartRouter(BaseLLMRouter):
    """
    A general-purpose smart router that can route between different types of agents
    based on task analysis and conversation context.
    """

    def __init__(
        self,
        routing_options: Dict[str, str],
        llm: Optional[BaseLLM] = None,
        context_description: Optional[str] = None,
        **kwargs,
    ):
        """
        Initialize the smart router.

        Args:
            routing_options: Dict mapping route names to descriptions
            llm: LLM instance for routing decisions
            context_description: Additional context about the workflow
            **kwargs: Additional arguments for BaseLLMRouter
        """
        super().__init__(llm=llm, **kwargs)
        self.routing_options = routing_options
        self.context_description = context_description or 'a multi-agent workflow'

    def get_routing_options(self) -> Dict[str, str]:
        return self.routing_options

    def get_routing_prompt(
        self,
        memory: BaseMemory,
        options: Dict[str, str],
        execution_context: dict = None,
    ) -> str:
        conversation = memory.get()

        # Format conversation history with smart truncation
        if isinstance(conversation, list):
            # Start with last message and add more if we have space
            messages = conversation[-5:]  # Last 5 messages
            conversation_text = self._truncate_conversation_for_tokens(messages)
        else:
            conversation_text = str(conversation)

        # Format options
        options_text = '\n'.join(
            [f'- {name}: {desc}' for name, desc in options.items()]
        )

        # Add execution context if available
        context_info = ''
        if execution_context:
            visit_counts = execution_context.get('node_visit_count', {})
            current_node = execution_context.get('current_node', 'unknown')
            iteration = execution_context.get('iteration_count', 0)

            # Create visit count warning if any node has been visited multiple times
            visit_warnings = []
            for node_name, count in visit_counts.items():
                if count >= 2:
                    visit_warnings.append(f'- {node_name}: visited {count} times')

            if visit_warnings:
                context_info = f"""

⚠️  EXECUTION CONTEXT (Avoid Infinite Loops):
Current iteration: {iteration}
Current node: {current_node}
Node visit counts:
{chr(10).join(visit_warnings)}

IMPORTANT: To prevent infinite loops, strongly prefer moving to different nodes or completing the workflow.
"""

        prompt = f"""You are a workflow coordinator for {self.context_description}.

Based on the conversation history below, decide which agent should handle the next step.
{context_info}
Available agents:
{options_text}

Recent conversation:
{conversation_text}

Instructions:
1. Analyze the conversation to understand what type of work is needed next
2. Choose the most appropriate agent from the available options
3. If you notice nodes being visited repeatedly, prefer different options or completion
4. Respond with ONLY the agent name (no explanations or additional text)

Agent to route to:"""

        return prompt

    def _truncate_conversation_for_tokens(
        self, messages: List[Any], max_tokens: int = 128000
    ) -> str:
        """
        Intelligently truncate conversation to fit within token limits.
        Prioritizes recent messages while ensuring we don't exceed token limits.
        """
        if not messages:
            return ''

        # Start with the most recent message
        truncated_messages = [messages[-1]]
        current_text = str(messages[-1])

        # Add older messages if we have space
        for msg in reversed(messages[:-1]):
            msg_text = str(msg)
            # Rough token estimation (4 chars per token is a common approximation)
            estimated_tokens = len(current_text + '\n' + msg_text) // 4

            if estimated_tokens <= max_tokens:
                truncated_messages.insert(0, msg)
                current_text = '\n'.join([str(m) for m in truncated_messages])
            else:
                break

        return '\n'.join([str(msg) for msg in truncated_messages])


class TaskClassifierRouter(BaseLLMRouter):
    """
    A router specialized for classifying tasks and routing to appropriate specialists.
    """

    def __init__(
        self,
        task_categories: Dict[str, Dict[str, Any]],
        llm: Optional[BaseLLM] = None,
        **kwargs,
    ):
        """
        Initialize the task classifier router.

        Args:
            task_categories: Dict mapping category names to their config:
                {
                    "category_name": {
                        "description": "What this category handles",
                        "keywords": ["keyword1", "keyword2"],  # Optional
                        "examples": ["example task 1", "example task 2"]  # Optional
                    }
                }
            llm: LLM instance for routing decisions
            **kwargs: Additional arguments for BaseLLMRouter
        """
        super().__init__(llm=llm, **kwargs)
        self.task_categories = task_categories

    def get_routing_options(self) -> Dict[str, str]:
        return {
            name: config['description'] for name, config in self.task_categories.items()
        }

    def get_routing_prompt(
        self,
        memory: BaseMemory,
        options: Dict[str, str],
        execution_context: dict = None,
    ) -> str:
        conversation = memory.get()

        # Get the latest user input or task
        if isinstance(conversation, list) and conversation:
            latest_task = str(conversation[-1])
        else:
            latest_task = str(conversation)

        # Build detailed category descriptions
        categories_detail = []
        for name, config in self.task_categories.items():
            detail = f"- {name}: {config['description']}"

            if 'keywords' in config:
                detail += f"\n  Keywords: {', '.join(config['keywords'])}"

            if 'examples' in config:
                detail += f"\n  Examples: {', '.join(config['examples'])}"

            categories_detail.append(detail)

        categories_text = '\n\n'.join(categories_detail)

        prompt = f"""You are a task classifier that routes requests to specialized agents.

Task to classify:
{latest_task}

Available categories:
{categories_text}

Instructions:
1. Analyze the task to understand what type of work it requires
2. Choose the most appropriate category from the available options
3. Consider keywords and examples to make the best match
4. Respond with ONLY the category name (no explanations)

Category:"""

        return prompt


class ReflectionRouter(BaseLLMRouter):
    """
    A router designed for reflection patterns like A -> B -> A -> C.
    Commonly used for main -> critic -> main -> final workflows where B is a reflection/critique step.
    Uses execution context to determine flow state and make intelligent routing decisions.
    """

    def __init__(
        self,
        flow_pattern: List[str],
        llm: Optional[BaseLLM] = None,
        allow_early_exit: bool = False,
        **kwargs,
    ):
        """
        Initialize the reflection router.

        Args:
            flow_pattern: List of node names defining the reflection pattern (e.g., ["main", "critic", "main", "final"])
            llm: LLM instance for routing decisions
            allow_early_exit: Whether to allow LLM to exit pattern early if appropriate
            **kwargs: Additional arguments for BaseLLMRouter
        """
        super().__init__(llm=llm, **kwargs)
        self.flow_pattern = flow_pattern
        self.allow_early_exit = allow_early_exit
        self.supports_self_reference = (
            True  # ReflectionRouter can return to the same node
        )

    def get_routing_options(self) -> Dict[str, str]:
        """Get available routing options based on flow pattern"""
        unique_nodes = list(
            dict.fromkeys(self.flow_pattern)
        )  # Preserve order, remove duplicates

        # Generate descriptions based on reflection pattern
        options = {}
        for node in unique_nodes:
            # Count occurrences and positions in pattern
            positions = [i for i, x in enumerate(self.flow_pattern) if x == node]
            if len(positions) > 1:
                options[node] = (
                    f"Step {positions} in the reflection pattern: {' -> '.join(self.flow_pattern)}"
                )
            else:
                options[node] = (
                    f"Step {positions[0] + 1} in the reflection pattern: {' -> '.join(self.flow_pattern)}"
                )

        return options

    def _get_next_step_in_pattern(self, execution_context: dict) -> Optional[str]:
        """Determine the next step in the reflection pattern based on execution context"""
        if not execution_context:
            return self.flow_pattern[0] if self.flow_pattern else None

        visit_counts = execution_context.get('node_visit_count', {})
        current_node = execution_context.get('current_node', '')

        # Find the current position in the pattern
        try:
            # Find where we are in the reflection pattern
            current_step = -1
            for i, node in enumerate(self.flow_pattern):
                node_visits = visit_counts.get(node, 0)

                # For nodes that appear multiple times, we need to track which occurrence
                if node == current_node:
                    # Count how many times this node should have been visited at this step
                    expected_visits = len(
                        [x for x in self.flow_pattern[: i + 1] if x == node]
                    )
                    if node_visits >= expected_visits:
                        current_step = i

            # Determine next step
            next_step_index = current_step + 1
            if next_step_index < len(self.flow_pattern):
                return self.flow_pattern[next_step_index]
            else:
                # Pattern completed
                return None

        except Exception:
            # Fallback to first step
            return self.flow_pattern[0] if self.flow_pattern else None

    def get_routing_prompt(
        self,
        memory: BaseMemory,
        options: Dict[str, str],
        execution_context: dict = None,
    ) -> str:
        conversation = memory.get()

        # Format conversation history
        if isinstance(conversation, list):
            conversation_text = '\n'.join(
                [str(msg) for msg in conversation[-3:]]
            )  # Last 3 messages for flow context
        else:
            conversation_text = str(conversation)

        # Determine suggested next step based on reflection pattern
        suggested_next = self._get_next_step_in_pattern(execution_context)

        # Format options
        options_text = '\n'.join(
            [f'- {name}: {desc}' for name, desc in options.items()]
        )

        # Add execution context info
        context_info = ''
        if execution_context:
            visit_counts = execution_context.get('node_visit_count', {})
            current_node = execution_context.get('current_node', 'unknown')
            iteration = execution_context.get('iteration_count', 0)

            # Show reflection pattern progress
            pattern_progress = []
            for i, node in enumerate(self.flow_pattern):
                visits = visit_counts.get(node, 0)
                expected_visits = len(
                    [x for x in self.flow_pattern[: i + 1] if x == node]
                )
                status = '✓' if visits >= expected_visits else '○'
                pattern_progress.append(f'{status} {node}')

            context_info = f"""
📋 REFLECTION PATTERN: {' → '.join(self.flow_pattern)}
📍 CURRENT PROGRESS: {' → '.join(pattern_progress)}
🎯 SUGGESTED NEXT: {suggested_next or 'Pattern Complete'}
💡 CURRENT NODE: {current_node} (iteration {iteration})
"""

        # Create prompt based on whether early exit is allowed
        if self.allow_early_exit:
            prompt = f"""You are a reflection coordinator managing this workflow pattern: {' → '.join(self.flow_pattern)}

{context_info}
Available options:
{options_text}

Recent conversation:
{conversation_text}

Instructions:
1. Follow the reflection pattern: {' → '.join(self.flow_pattern)}
2. The suggested next step is: {suggested_next or 'Pattern Complete'}
3. You may exit early if the reflection cycle is complete
4. Consider conversation context and reflection progress
5. Respond with ONLY the agent name (no explanations)

Next agent:"""
        else:
            prompt = f"""You are a reflection coordinator managing this strict reflection pattern: {' → '.join(self.flow_pattern)}

{context_info}
Available options:
{options_text}

Recent conversation:
{conversation_text}

Instructions:
1. STRICTLY follow the reflection pattern: {' → '.join(self.flow_pattern)}
2. The next step should be: {suggested_next or 'Pattern Complete'}
3. Do not deviate from the pattern unless absolutely necessary
4. Respond with ONLY the agent name (no explanations)

Next agent:"""

        return prompt


class PlanExecuteRouter(BaseLLMRouter):
    """
    A router that implements plan-and-execute patterns like Cursor.
    Creates execution plans and routes through steps sequentially.
    """

    def __init__(
        self,
        agents: Dict[str, str],  # agent_name -> description mapping
        planner_agent: str = 'planner',
        executor_agent: str = 'executor',
        reviewer_agent: Optional[str] = None,
        llm: Optional[BaseLLM] = None,
        max_retries: int = 3,
        **kwargs,
    ):
        """
        Initialize the plan-execute router.

        Args:
            agents: Dict mapping agent names to their descriptions/capabilities
            planner_agent: Name of the agent responsible for creating plans
            executor_agent: Name of the agent responsible for executing steps
            reviewer_agent: Optional name of the agent responsible for reviewing results
            llm: LLM instance for routing decisions
            max_retries: Maximum retries for step execution
            **kwargs: Additional arguments for BaseLLMRouter
        """
        super().__init__(llm=llm, **kwargs)
        self.agents = agents
        self.planner_agent = planner_agent
        self.executor_agent = executor_agent
        self.reviewer_agent = reviewer_agent
        self.max_retries = max_retries
        self.supports_self_reference = (
            True  # Can route to same agent for iterative execution
        )

    def get_routing_options(self) -> Dict[str, str]:
        """Get available routing options based on configured agents"""
        return self.agents

    def get_routing_prompt(
        self,
        memory: BaseMemory,
        options: Dict[str, str],
        execution_context: dict = None,
    ) -> str:
        conversation = memory.get()

        # Format conversation history
        if isinstance(conversation, list):
            conversation_text = '\n'.join(
                [str(msg) for msg in conversation[-3:]]
            )  # Last 3 messages for context
        else:
            conversation_text = str(conversation)

        # Check if we have a plan in memory
        current_plan = (
            memory.get_current_plan() if hasattr(memory, 'get_current_plan') else None
        )

        if current_plan is None:
            # No plan exists - route to planner
            return self._create_planning_prompt(conversation_text, options)
        else:
            # Plan exists - determine next action based on plan state
            return self._create_execution_prompt(
                current_plan, conversation_text, options, execution_context
            )

    def _create_planning_prompt(
        self, conversation_text: str, options: Dict[str, str]
    ) -> str:
        """Create prompt for initial planning phase"""
        options_text = '\n'.join(
            [f'- {name}: {desc}' for name, desc in options.items()]
        )

        prompt = f"""You are coordinating a plan-and-execute workflow. No execution plan exists yet.

Available agents:
{options_text}

Recent conversation:
{conversation_text}

TASK: Create an execution plan by routing to the {self.planner_agent}.

Instructions:
1. Route to "{self.planner_agent}" to create a detailed execution plan
2. The planner will break down the task into sequential steps
3. Each step will specify which agent should execute it
4. Respond with ONLY the agent name: {self.planner_agent}

Next agent:"""

        return prompt

    def _create_execution_prompt(
        self,
        plan: ExecutionPlan,
        conversation_text: str,
        options: Dict[str, str],
        execution_context: dict = None,
    ) -> str:
        """Create prompt for execution phase based on current plan state"""

        # Get next steps that are ready to execute
        next_steps = plan.get_next_steps()

        # Format plan progress
        progress_lines = []
        for step in plan.steps:
            status_icon = {
                StepStatus.PENDING: '○',
                StepStatus.IN_PROGRESS: '⏳',
                StepStatus.COMPLETED: '✅',
                StepStatus.FAILED: '❌',
                StepStatus.SKIPPED: '⏭️',
            }.get(step.status, '○')
            progress_lines.append(
                f'{status_icon} {step.id}: {step.description} (→ {step.agent})'
            )

        progress_text = '\n'.join(progress_lines)

        # Determine what to do next
        if plan.is_completed():
            # All steps completed
            if self.reviewer_agent and self.reviewer_agent in options:
                action = f'Route to {self.reviewer_agent} for final review'
                suggested_agent = self.reviewer_agent
            else:
                action = 'Plan completed - route to any agent for final output'
                suggested_agent = next(iter(options.keys()))  # First available agent
        elif plan.has_failed_steps():
            # Some steps failed - need recovery
            failed_steps = [
                step for step in plan.steps if step.status == StepStatus.FAILED
            ]
            failed_step = failed_steps[0]  # Focus on first failed step
            action = f"Handle failed step '{failed_step.id}' - route to {failed_step.agent} for retry"
            suggested_agent = failed_step.agent
        elif next_steps:
            # There are steps ready to execute
            next_step = next_steps[0]  # Execute first ready step
            action = f"Execute step '{next_step.id}' - route to {next_step.agent}"
            suggested_agent = next_step.agent
        else:
            # Waiting for dependencies
            action = f'Waiting for dependencies - route to {self.executor_agent} for status check'
            suggested_agent = self.executor_agent

        options_text = '\n'.join(
            [f'- {name}: {desc}' for name, desc in options.items()]
        )

        # Add execution context info
        context_info = ''
        if execution_context:
            current_node = execution_context.get('current_node', 'unknown')
            iteration = execution_context.get('iteration_count', 0)

            context_info = f"""
💡 EXECUTION CONTEXT:
Current node: {current_node} (iteration {iteration})
"""

        prompt = f"""You are coordinating plan execution in a plan-and-execute workflow.

📋 EXECUTION PLAN: {plan.title}
{plan.description}

📊 CURRENT PROGRESS:
{progress_text}

🎯 NEXT ACTION: {action}
🎯 SUGGESTED AGENT: {suggested_agent}
{context_info}
Available agents:
{options_text}

Recent conversation:
{conversation_text}

Instructions:
1. Follow the execution plan step by step
2. Route to the suggested agent: {suggested_agent}
3. Each agent will execute their assigned step
4. Continue until all steps are completed
5. Respond with ONLY the agent name (no explanations)

Next agent:"""

        return prompt


class ConversationAnalysisRouter(BaseLLMRouter):
    """
    A router that analyzes conversation flow and context to make routing decisions.
    """

    def __init__(
        self,
        routing_logic: Dict[str, str],
        analysis_depth: int = 3,
        llm: Optional[BaseLLM] = None,
        **kwargs,
    ):
        """
        Initialize the conversation analysis router.

        Args:
            routing_logic: Dict mapping route names to routing criteria
            analysis_depth: Number of recent messages to analyze
            llm: LLM instance for routing decisions
            **kwargs: Additional arguments for BaseLLMRouter
        """
        super().__init__(llm=llm, **kwargs)
        self.routing_logic = routing_logic
        self.analysis_depth = analysis_depth

    def get_routing_options(self) -> Dict[str, str]:
        return self.routing_logic

    def get_routing_prompt(
        self,
        memory: BaseMemory,
        options: Dict[str, str],
        execution_context: dict = None,
    ) -> str:
        conversation = memory.get()

        # Analyze recent conversation
        if isinstance(conversation, list):
            recent_messages = conversation[-self.analysis_depth :]
            conversation_text = '\n'.join(
                [f'Message {i+1}: {msg}' for i, msg in enumerate(recent_messages)]
            )
        else:
            conversation_text = str(conversation)

        # Format routing logic
        logic_text = '\n'.join(
            [f'- {name}: {criteria}' for name, criteria in options.items()]
        )

        # Add execution context if available
        context_info = ''
        if execution_context:
            visit_counts = execution_context.get('node_visit_count', {})
            current_node = execution_context.get('current_node', 'unknown')
            iteration = execution_context.get('iteration_count', 0)

            # Create visit count warning if any node has been visited multiple times
            visit_warnings = []
            for node_name, count in visit_counts.items():
                if count >= 2:
                    visit_warnings.append(f'- {node_name}: visited {count} times')

            if visit_warnings:
                context_info = f"""

⚠️  EXECUTION CONTEXT (Avoid Infinite Loops):
Current iteration: {iteration}
Current node: {current_node}
Node visit counts:
{chr(10).join(visit_warnings)}

CRITICAL: Excessive node revisits detected! Consider completing the workflow or choosing different paths.
"""

        prompt = f"""You are a conversation flow analyzer that determines the next step in a workflow.

Recent conversation (last {self.analysis_depth} messages):
{conversation_text}
{context_info}
Routing logic:
{logic_text}

Instructions:
1. Analyze the conversation flow and current state
2. Consider what has been accomplished and what needs to happen next
3. If nodes are being revisited too frequently, strongly prefer completion or different paths
4. Choose the route that best matches the current conversation state
5. Respond with ONLY the route name (no explanations)

Next route:"""

        return prompt


def create_llm_router(router_type: str, **config) -> Callable[[BaseMemory], str]:
    """
    Factory function to create LLM-powered routers with different configurations.

    Args:
        router_type: Type of router ("smart", "task_classifier", "conversation_analysis", "reflection", "plan_execute")
        **config: Configuration specific to the router type

    Returns:
        Callable router function that can be used in Arium workflows

    Examples:
        # Smart router
        router = create_llm_router(
            "smart",
            routing_options={
                "researcher": "Gather information and conduct research",
                "analyst": "Analyze data and perform calculations",
                "writer": "Create reports and summaries"
            }
        )

        # Task classifier router
        router = create_llm_router(
            "task_classifier",
            task_categories={
                "research": {
                    "description": "Research and information gathering tasks",
                    "keywords": ["research", "find", "investigate", "gather"],
                    "examples": ["Find information about...", "Research the topic of..."]
                },
                "analysis": {
                    "description": "Data analysis and computational tasks",
                    "keywords": ["analyze", "calculate", "compute", "data"],
                    "examples": ["Analyze the data...", "Calculate the..."]
                }
            }
        )

        # Reflection router for A -> B -> A -> C patterns
        router = create_llm_router(
            "reflection",
            flow_pattern=["main_agent", "critic", "main_agent", "final_agent"],
            allow_early_exit=False
        )

        # Plan-Execute router for Cursor-style workflows
        router = create_llm_router(
            "plan_execute",
            agents={
                "planner": "Creates detailed execution plans",
                "developer": "Implements code and features",
                "tester": "Tests implementations",
                "reviewer": "Reviews final results"
            },
            planner_agent="planner",
            executor_agent="developer"
        )
    """

    if router_type == 'smart':
        if 'routing_options' not in config:
            raise ValueError("SmartRouter requires 'routing_options' parameter")

        router_instance = SmartRouter(**config)

    elif router_type == 'task_classifier':
        if 'task_categories' not in config:
            raise ValueError(
                "TaskClassifierRouter requires 'task_categories' parameter"
            )

        router_instance = TaskClassifierRouter(**config)

    elif router_type == 'conversation_analysis':
        if 'routing_logic' not in config:
            raise ValueError(
                "ConversationAnalysisRouter requires 'routing_logic' parameter"
            )

        router_instance = ConversationAnalysisRouter(**config)

    elif router_type == 'reflection':
        if 'flow_pattern' not in config:
            raise ValueError("ReflectionRouter requires 'flow_pattern' parameter")

        router_instance = ReflectionRouter(**config)

    elif router_type == 'plan_execute':
        if 'agents' not in config:
            raise ValueError("PlanExecuteRouter requires 'agents' parameter")

        router_instance = PlanExecuteRouter(**config)

    else:
        raise ValueError(f'Unknown router type: {router_type}')

    # Get the routing options for type annotation
    options = router_instance.get_routing_options()
    option_names = tuple(options.keys())

    # Create proper Literal type for validation
    from typing import Literal

    if len(option_names) == 1:
        # Handle single option case
        literal_type = Literal[option_names[0]]
    else:
        # Handle multiple options case
        literal_type = Literal[option_names]

    # Return a function that can be used as a router
    async def router_function(memory: BaseMemory, execution_context: dict = None):
        """Generated router function that uses LLM for routing decisions"""
        return await router_instance.route(memory, execution_context)

    # Add proper type annotations for validation
    router_function.__annotations__ = {'memory': BaseMemory, 'return': literal_type}

    # Transfer router instance attributes to the function for validation
    router_function.supports_self_reference = getattr(
        router_instance, 'supports_self_reference', False
    )

    return router_function


def llm_router(
    routing_options: Dict[str, str],
    llm: Optional[BaseLLM] = None,
    context_description: Optional[str] = None,
    **kwargs,
):
    """
    Decorator to create LLM-powered routers with a simple interface.

    Args:
        routing_options: Dict mapping route names to descriptions
        llm: LLM instance for routing decisions
        context_description: Additional context about the workflow
        **kwargs: Additional router configuration

    Returns:
        Decorator function

    Example:
        @llm_router({
            "researcher": "Gather information and conduct research",
            "analyst": "Analyze data and perform calculations",
            "writer": "Create reports and summaries"
        })
        def my_smart_router(memory: BaseMemory) -> Literal["researcher", "analyst", "writer"]:
            pass  # Implementation is provided by decorator
    """

    def decorator(func):
        # Extract return type annotation to validate routing options
        if hasattr(func, '__annotations__') and 'return' in func.__annotations__:
            return_annotation = func.__annotations__['return']

            # Validate that routing options match the return type
            if (
                hasattr(return_annotation, '__origin__')
                and return_annotation.__origin__ is Union
            ):
                # Handle Literal types
                args = get_args(return_annotation)
                if args and hasattr(args[0], '__origin__'):
                    literal_values = get_args(args[0])
                    option_keys = set(routing_options.keys())
                    literal_set = set(literal_values)

                    if option_keys != literal_set:
                        logger.warning(
                            f"Routing options {option_keys} don't match return type {literal_set}"
                        )

        # Create the router instance
        router_instance = SmartRouter(
            routing_options=routing_options,
            llm=llm,
            context_description=context_description,
            **kwargs,
        )

        @wraps(func)
        async def wrapper(memory: BaseMemory, execution_context: dict = None):
            return await router_instance.route(memory, execution_context)

        # Preserve the original function's type annotations including return type
        wrapper.__annotations__ = func.__annotations__.copy()

        # Ensure the return annotation is properly set
        if 'return' in func.__annotations__:
            wrapper.__annotations__['return'] = func.__annotations__['return']

        return wrapper

    return decorator


# Convenience functions for common routing patterns


def create_research_analysis_router(
    research_agent: str = 'researcher',
    analysis_agent: str = 'analyst',
    summary_agent: str = 'summarizer',
    llm: Optional[BaseLLM] = None,
) -> Callable[[BaseMemory], str]:
    """
    Create a router for common research -> analysis -> summary workflows.

    Args:
        research_agent: Name of the research agent
        analysis_agent: Name of the analysis agent
        summary_agent: Name of the summary agent
        llm: LLM instance for routing decisions

    Returns:
        Router function for research/analysis workflows
    """
    return create_llm_router(
        'task_classifier',
        task_categories={
            research_agent: {
                'description': 'Research and information gathering tasks',
                'keywords': [
                    'research',
                    'find',
                    'investigate',
                    'gather',
                    'search',
                    'lookup',
                ],
                'examples': [
                    'Find information about...',
                    'Research the topic of...',
                    'Gather data on...',
                ],
            },
            analysis_agent: {
                'description': 'Data analysis and computational tasks',
                'keywords': [
                    'analyze',
                    'calculate',
                    'compute',
                    'data',
                    'numbers',
                    'statistics',
                ],
                'examples': [
                    'Analyze the data...',
                    'Calculate the...',
                    'Compute statistics for...',
                ],
            },
            summary_agent: {
                'description': 'Summarization and conclusion tasks',
                'keywords': [
                    'summarize',
                    'conclude',
                    'report',
                    'final',
                    'overview',
                    'wrap up',
                ],
                'examples': [
                    'Summarize the findings...',
                    'Create a final report...',
                    'Conclude the analysis...',
                ],
            },
        },
        llm=llm,
    )


def create_main_critic_reflection_router(
    main_agent: str = 'main_agent',
    critic_agent: str = 'critic',
    final_agent: str = 'final_agent',
    allow_early_exit: bool = False,
    llm: Optional[BaseLLM] = None,
) -> Callable[[BaseMemory], str]:
    """
    Create a router for the A -> B -> A -> C reflection pattern (main -> critic -> main -> final).

    Args:
        main_agent: Name of the main agent (appears twice in pattern)
        critic_agent: Name of the critic agent for reflection
        final_agent: Name of the final agent
        allow_early_exit: Whether to allow LLM to exit reflection early if appropriate
        llm: LLM instance for routing decisions

    Returns:
        Router function for main/critic/final reflection workflows
    """
    return create_llm_router(
        'reflection',
        flow_pattern=[main_agent, critic_agent, main_agent, final_agent],
        allow_early_exit=allow_early_exit,
        llm=llm,
    )


def create_plan_execute_router(
    planner_agent: str = 'planner',
    executor_agent: str = 'executor',
    reviewer_agent: Optional[str] = None,
    additional_agents: Optional[Dict[str, str]] = None,
    llm: Optional[BaseLLM] = None,
) -> Callable[[BaseMemory], str]:
    """
    Create a router for plan-and-execute workflows like Cursor.

    Args:
        planner_agent: Name of the agent responsible for creating plans
        executor_agent: Name of the agent responsible for executing steps
        reviewer_agent: Optional name of the agent responsible for reviewing results
        additional_agents: Additional agents that can be used in execution steps
        llm: LLM instance for routing decisions

    Returns:
        Router function for plan-execute workflows
    """
    agents = {
        planner_agent: 'Creates detailed execution plans by breaking down tasks into sequential steps',
        executor_agent: 'Executes individual steps from the execution plan',
    }

    if reviewer_agent:
        agents[reviewer_agent] = 'Reviews and validates completed work'

    if additional_agents:
        agents.update(additional_agents)

    return create_llm_router(
        'plan_execute',
        agents=agents,
        planner_agent=planner_agent,
        executor_agent=executor_agent,
        reviewer_agent=reviewer_agent,
        llm=llm,
    )


# Backward compatibility alias
def create_main_critic_flow_router(
    main_agent: str = 'main_agent',
    critic_agent: str = 'critic',
    final_agent: str = 'final_agent',
    allow_early_exit: bool = False,
    llm: Optional[BaseLLM] = None,
) -> Callable[[BaseMemory], str]:
    """
    DEPRECATED: Use create_main_critic_reflection_router instead.
    Create a router for the A -> B -> A -> C reflection pattern (main -> critic -> main -> final).
    """
    return create_main_critic_reflection_router(
        main_agent=main_agent,
        critic_agent=critic_agent,
        final_agent=final_agent,
        allow_early_exit=allow_early_exit,
        llm=llm,
    )



from abc import ABC, abstractmethod
from typing import TypeVar, Generic, List, Dict, Optional, Any
from dataclasses import dataclass, field
from enum import Enum

# Define the generic type variable
T = TypeVar('T')


class StepStatus(Enum):
    """Status of a plan step"""

    PENDING = 'pending'
    IN_PROGRESS = 'in_progress'
    COMPLETED = 'completed'
    FAILED = 'failed'
    SKIPPED = 'skipped'


@dataclass
class PlanStep:
    """Represents a single step in an execution plan"""

    id: str
    description: str
    agent: str  # Which agent should execute this step
    dependencies: List[str] = field(default_factory=list)  # Step IDs this depends on
    status: StepStatus = StepStatus.PENDING
    result: Optional[str] = None
    error: Optional[str] = None
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class ExecutionPlan:
    """Represents a complete execution plan"""

    id: str
    title: str
    description: str
    steps: List[PlanStep] = field(default_factory=list)
    created_by: str = 'planner'
    status: str = 'active'  # active, completed, failed, paused
    metadata: Dict[str, Any] = field(default_factory=dict)

    def get_next_steps(self) -> List[PlanStep]:
        """Get steps that are ready to execute (pending with no pending dependencies)"""
        next_steps = []
        for step in self.steps:
            if step.status == StepStatus.PENDING:
                # Check if all dependencies are completed
                if all(
                    self.get_step(dep_id).status == StepStatus.COMPLETED
                    for dep_id in step.dependencies
                ):
                    next_steps.append(step)
        return next_steps

    def get_step(self, step_id: str) -> Optional[PlanStep]:
        """Get a step by ID"""
        for step in self.steps:
            if step.id == step_id:
                return step
        return None

    def mark_step_completed(self, step_id: str, result: str = None):
        """Mark a step as completed"""
        step = self.get_step(step_id)
        if step:
            step.status = StepStatus.COMPLETED
            step.result = result

    def mark_step_failed(self, step_id: str, error: str = None):
        """Mark a step as failed"""
        step = self.get_step(step_id)
        if step:
            step.status = StepStatus.FAILED
            step.error = error

    def is_completed(self) -> bool:
        """Check if all steps are completed"""
        return all(step.status == StepStatus.COMPLETED for step in self.steps)

    def has_failed_steps(self) -> bool:
        """Check if any steps have failed"""
        return any(step.status == StepStatus.FAILED for step in self.steps)


class BaseMemory(ABC, Generic[T]):
    @abstractmethod
    def add(self, m: T):
        pass

    @abstractmethod
    def get(self) -> List[T]:
        pass

    # Plan management methods (optional - only implemented by memory classes that support plans)
    def add_plan(self, plan: ExecutionPlan):
        """Add an execution plan (override in subclasses that support plans)"""
        raise NotImplementedError('This memory type does not support plans')

    def get_current_plan(self) -> Optional[ExecutionPlan]:
        """Get the current active plan (override in subclasses that support plans)"""
        return None

    def update_plan(self, plan: ExecutionPlan):
        """Update an existing plan (override in subclasses that support plans)"""
        raise NotImplementedError('This memory type does not support plans')

    def get_plan(self, plan_id: str) -> Optional[ExecutionPlan]:
        """Get a plan by ID (override in subclasses that support plans)"""
        return None


class MessageMemory(BaseMemory[Dict[str, str]]):
    def __init__(self):
        self.messages = []

    def add(self, message: Dict[str, str]):
        self.messages.append(message)

    def get(self) -> List[Dict[str, str]]:
        return self.messages


class PlanAwareMemory(BaseMemory[Dict[str, str]]):
    """Enhanced memory that supports both messages and execution plans"""

    def __init__(self):
        self.messages = []
        self.plans: Dict[str, ExecutionPlan] = {}
        self.current_plan_id: Optional[str] = None

    def add(self, message: Dict[str, str]):
        """Add a message to memory"""
        self.messages.append(message)

    def get(self) -> List[Dict[str, str]]:
        """Get all messages"""
        return self.messages

    # Plan management methods
    def add_plan(self, plan: ExecutionPlan):
        """Add an execution plan and set it as current"""
        self.plans[plan.id] = plan
        self.current_plan_id = plan.id

    def get_current_plan(self) -> Optional[ExecutionPlan]:
        """Get the current active plan"""
        if self.current_plan_id and self.current_plan_id in self.plans:
            return self.plans[self.current_plan_id]
        return None

    def update_plan(self, plan: ExecutionPlan):
        """Update an existing plan"""
        if plan.id in self.plans:
            self.plans[plan.id] = plan
        else:
            raise ValueError(f'Plan {plan.id} not found in memory')

    def get_plan(self, plan_id: str) -> Optional[ExecutionPlan]:
        """Get a plan by ID"""
        return self.plans.get(plan_id)

    def set_current_plan(self, plan_id: str):
        """Set the current active plan"""
        if plan_id in self.plans:
            self.current_plan_id = plan_id
        else:
            raise ValueError(f'Plan {plan_id} not found in memory')

    def get_all_plans(self) -> List[ExecutionPlan]:
        """Get all plans"""
        return list(self.plans.values())

    def remove_plan(self, plan_id: str):
        """Remove a plan from memory"""
        if plan_id in self.plans:
            del self.plans[plan_id]
            if self.current_plan_id == plan_id:
                self.current_plan_id = None
